import astropy.units as u
import nidaqmx
import numpy as np
import pytest

from openwfs.devices import is_loaded

if not is_loaded(nidaqmx):
    pytest.skip(nidaqmx.messagea, allow_module_level=True)

from openwfs.devices import ScanningMicroscope, Axis
from openwfs.devices.galvo_scanner import InputChannel
from openwfs.utilities import coordinate_range


@pytest.mark.parametrize("start, stop", [(0.0, 1.0), (1.0, 0.0)])
def test_scan_axis(start, stop):
    """Tests if the Axis class generates the correct voltage sequences for stepping and scanning."""
    maximum_acceleration = 1 * u.V / u.ms**2
    scale = 440 * u.um / u.V
    v_min = -1.0 * u.V
    v_max = 2.0 * u.V
    a = Axis(
        channel="Dev4/ao0",
        v_min=v_min,
        v_max=v_max,
        maximum_acceleration=maximum_acceleration,
        scale=scale,
    )
    assert a.channel == "Dev4/ao0"
    assert a.v_min == v_min
    assert a.v_max == v_max
    assert a.maximum_acceleration == maximum_acceleration
    assert a.to_volt(0.0) == v_min
    assert a.to_volt(1.0) == v_max
    assert a.to_volt(2.0) == v_max  # test clipping
    assert a.to_volt(-0.1) == v_min  # test clipping

    # test step
    sample_rate = 0.5 * u.MHz
    step = a.step(start, stop, sample_rate)
    # plt.plot(step)
    # plt.show()
    assert np.isclose(step[0], -1.0 * u.V if start == 0.0 else 2.0 * u.V)
    assert np.isclose(step[-1], 2.0 * u.V if start == 0.0 else -1.0 * u.V)
    assert np.all(step >= v_min)
    assert np.all(step <= v_max)
    acceleration = np.diff(np.diff(step)) * sample_rate**2
    assert np.all(np.abs(acceleration) <= maximum_acceleration * 1.01)
    center = 0.5 * (start + stop)
    amplitude = 0.5 * (stop - start)

    # test clipping
    assert np.allclose(step, a.step(center - 1.1 * amplitude, center + 1.1 * amplitude, sample_rate))

    # test scan. Note that we cannot use the full scan range because we need
    # some time to accelerate / decelerate
    sample_count = 10000
    scan, launch, land, linear_region = a.scan(
        center - 0.8 * amplitude, center + 0.8 * amplitude, sample_count, sample_rate
    )
    half_pixel = 0.8 * amplitude / sample_count
    # plt.plot(scan)
    # plt.show()
    assert linear_region.stop - linear_region.start == sample_count
    assert linear_region.start == len(scan) - linear_region.stop
    assert np.isclose(scan[0], a.to_volt(launch))
    assert np.isclose(scan[-1], a.to_volt(land))
    assert np.isclose(scan[linear_region.start], a.to_volt(center - 0.8 * amplitude + half_pixel))
    assert np.isclose(scan[linear_region.stop - 1], a.to_volt(center + 0.8 * amplitude - half_pixel))
    speed = np.diff(scan[linear_region])
    assert np.allclose(speed, speed[0])  # speed should be constant

    acceleration = np.diff(np.diff(scan)) * sample_rate**2
    assert np.all(np.abs(acceleration) <= maximum_acceleration * 1.01)


def make_scanner(bidirectional, direction, reference_zoom):
    scale = 440 * u.um / u.V
    sample_rate = 0.5 * u.MHz
    input_channel = InputChannel(channel="Dev4/ai0", v_min=-1.0 * u.V, v_max=1.0 * u.V)
    y_axis = Axis(
        channel="Dev4/ao0",
        v_min=-2.0 * u.V,
        v_max=2.0 * u.V,
        maximum_acceleration=10 * u.V / u.ms**2,
        scale=scale,
    )
    x_axis = Axis(
        channel="Dev4/ao1",
        v_min=-2.0 * u.V,
        v_max=2.0 * u.V,
        maximum_acceleration=10 * u.V / u.ms**2,
        scale=scale,
    )
    return ScanningMicroscope(
        bidirectional=bidirectional,
        sample_rate=sample_rate,
        resolution=1024,
        input=input_channel,
        y_axis=y_axis,
        x_axis=x_axis,
        test_pattern=direction,
        reference_zoom=reference_zoom,
    )


@pytest.mark.parametrize("direction", ["horizontal", "vertical"])
@pytest.mark.parametrize("bidirectional", [False, True])
def test_scan_pattern(direction, bidirectional):
    """A unit test for scanning patterns."""
    reference_zoom = 1.2
    scanner = make_scanner(bidirectional, direction, reference_zoom)
    assert np.allclose(scanner.extent, scanner._x_axis.scale * 4.0 * u.V / reference_zoom)
    # plt.imshow(scanner.read())
    # plt.show()

    # check if returned pattern is correct
    (y, x) = coordinate_range(
        (scanner._resolution, scanner._resolution),
        10000 / reference_zoom,
        offset=(5000, 5000),
    )
    full = scanner.read().astype("float32") - 0x8000

    pixel_size = full[1, 1] - full[0, 0]

    if direction == "horizontal":
        assert np.allclose(full, full[0, :])  # all rows should be the same
        assert np.allclose(x, full, atol=0.2 * pixel_size)  # some rounding due to quantization
    else:
        # all columns should be the same (note we need to keep the last dimension for correct broadcasting)
        assert np.allclose(full, full[:, 0:1])
        assert np.allclose(y, full, atol=0.2 * pixel_size)

    # test setting the ROI
    left = 10
    top = 30
    width = 29
    height = 17
    scanner.left = left
    scanner.top = top
    scanner.width = width
    scanner.height = height

    assert scanner.left == left
    assert scanner.top == top
    assert scanner.width == width
    assert scanner.height == height
    assert scanner.data_shape == (height, width)

    roi = scanner.read().astype("float32") - 0x8000
    assert np.allclose(full[top : (top + height), left : (left + width)], roi, atol=0.2 * pixel_size)


@pytest.mark.parametrize("bidirectional", [False, True])
def test_park_beam(bidirectional):
    """A unit test for parking the beam of a DAQ scanner."""
    reference_zoom = 1.2
    scanner = make_scanner(bidirectional, "horizontal", reference_zoom)

    # Park beam horizontally
    scanner.top = 3
    scanner.left = 4
    scanner.width = 1
    scanner.height = 2

    img = scanner.read()
    assert img.shape == (2, 1)
    voltages = scanner._scan_pattern
    assert np.allclose(voltages[1, :], voltages[1, 0])  # all voltages should be the same

    # Park beam vertically
    scanner.width = 2
    scanner.height = 1
    img = scanner.read()
    assert img.shape == (1, 2)
    voltages = scanner._scan_pattern
    assert np.allclose(voltages[0, :], voltages[0, 0])

    # Park beam horizontally and vertically
    scanner.width = 1
    scanner.height = 1
    img = scanner.read()
    assert img.shape == (1, 1)
    voltages = scanner._scan_pattern
    assert np.allclose(voltages[1, :], voltages[1, 0])  # all voltages should be the same
    assert np.allclose(voltages[0, :], voltages[0, 0])  # all voltages should be the same

    # TODO: add test for zooming



import warnings
from typing import Optional

import astropy.units as u
import numpy as np
from astropy.units import Quantity
from harvesters.core import Harvester

from ..core import Detector


class Camera(Detector):
    """Adapter for GenICam/GenTL cameras.

    Attributes:
        _nodes: The GenICam node map of the camera.
            This map can be used to access camera properties,
            see the `GenICam/GenAPI documentation <https://www.emva.org/standards-technology/genicam/>`_
            and the `Standard Features Naming Convention
            <https://www.emva.org/wp-content/uploads/GenICam_SFNC_2_3.pdf>`_ for more details.

            The node map should not be used to set properties that are available as properties in the Camera object,
            such as ``exposure``, ``width``, ``height``, ``binning``, etc.

            Also, the node map should not be used to set properties while the camera is fetching a frame (i.e.,
            between ``trigger()`` and calling ``result()`` on the returned concurrent.futures.Future object).

    Note:
        This class is a thin wrapper around the Harvesters module,
        which is a generic adapter for GenICam/GenTL cameras.

    Example:
        >>> camera = Camera(cti_file=R"C:\\Program Files\\Basler\\pylon 7\\Runtime\\x64\\ProducerU3V.cti")
        >>> camera.exposure = 10 * u.ms
        >>> frame = camera.read()
    """

    __slots__ = ("_harvester", "_camera", "_nodes")

    def __init__(
        self,
        cti_file: str,
        serial_number: Optional[str] = None,
        multi_threaded=True,
        **kwargs,
    ):
        """
        Initialize the Camera object.

        By default, the camera is configured to use the highest monochromatic bitdepth available.
        The trigger mode is set to `Software`, and `ExposureMode` to `Timed` (`ExposureAuto` is `Off`).
        The ROI is reset to the full sensor.
        To change any of these options, pass them as keyword arguments to the constructor, e. g. `PixelFormat='RGB16'`

        Args:
            cti_file: The path to the GenTL producer file.
                This path depends on where the driver for the camera is installed.
                For Basler cameras, this is typically located in
                R"C:\\Program Files\\Basler\\pylon 7\\Runtime\\x64\\ProducerU3V.cti".

            serial_number: The serial number of the camera.
                When omitted, the first camera found is selected.
            **kwargs: Additional keyword arguments.
                These arguments are transferred to the node map of the camera. They must follow the `genicam` standard.
        """
        self._harvester = Harvester()

        try:
            # Try to add the GenTL producer file (cti_file)
            self._harvester.add_file(cti_file, check_validity=True)
            print(f"Successfully loaded CTI file: {cti_file}")
        except Exception as e:
            # Catch any errors during the file loading process and provide a user-friendly message
            print(f"Failed to load CTI file: {cti_file}")
            print(f"Error: {str(e)}")
            print(
                "Please ensure that the CTI file exists at the specified location "
                "and that it is a valid GenTL producer file. You can download or "
                "locate the file from the camera manufacturer's website or SDK, "
                "such as the Basler pylon SDK."
            )
            raise

        self._harvester.update()

        # open the camera, use the serial_number to select the camera if it is specified.
        search_key = {"serial_number": serial_number} if serial_number is not None else None
        self._camera = self._harvester.create(search_key=search_key)
        nodes = self._camera.remote_device.node_map
        self._nodes = nodes

        # print(dir(nodes))  # for debugging, should go in a separate function

        # set triggering to 'Software', so that we can trigger the camera by calling `trigger`.
        # turn off auto exposure so that `duration` accurately reflects the required measurement time.
        # note: these default settings can be overridden by the kwargs.
        settings = {
            "TriggerMode": "On",
            "TriggerSource": "Software",
            "ExposureMode": "Timed",
            "ExposureAuto": "Off",
            "BinningHorizontal": 1,
            "BinningVertical": 1,
            "OffsetX": 0,
            "OffsetY": 0,
            "Width": nodes.Width.max,
            "Height": nodes.Height.max,
        }

        # by default use the highest bitdepth
        if not "PixelFormat" in kwargs.keys():
            for mode in [
                "Mono16",
                "Mono14p",  # Mono14Packed does not exist
                "Mono14",
                "Mono12Packed",
                "Mono12p",
                "Mono12",
                "Mono10Packed",
                "Mono10p",
                "Mono10",
                "Mono8",
            ]:
                try:
                    nodes.PixelFormat.value = mode
                    break
                except Exception as e:
                    continue

        # set additional properties specified in the kwargs
        settings.update(kwargs)

        for key, value in settings.items():
            try:
                getattr(nodes, key).value = value
            except AttributeError:
                print(f"Warning: could not set camera property {key} to {value}")

        #  Todo:
        #         automatically expose a selection of properties in the node map as
        #         properties of the Camera object.
        #

        try:
            pixel_size = [
                nodes.SensorPixelHeight.value,
                nodes.SensorPixelWidth.value,
            ] * u.um
        except AttributeError:  # the SensorPixelWidth feature is optional
            pixel_size = None

        super().__init__(
            multi_threaded=multi_threaded,
            data_shape=None,
            pixel_size=pixel_size,
            duration=None,
            latency=0.0 * u.ms,
        )
        self._camera.start()

    def __del__(self):
        if hasattr(self, "_camera"):
            self._camera.stop()
            self._camera.destroy()
        if hasattr(self, "_harvester"):
            self._harvester.reset()

    def _do_trigger(self):
        """Executes the software trigger command on the camera.

        This method is called by the trigger() method to initiate image acquisition.
        """
        self._nodes.TriggerSoftware.execute()

    def paused(self):
        """Returns a context manager for pausing the camera.

        Usage ::
            >>> with camera.paused():
            >>>     camera.nodes.SomeNode.value = 10
        """
        return _CameraPause(self._camera)

    def _fetch(self, *args, **kwargs) -> np.ndarray:
        """Fetches a frame from the camera.

        This method is called by the trigger() method (typically on a worker thread)
        to retrieve the image data after the camera has been triggered.
        Returns:
            np.ndarray: The image data from the camera.

        Raises:
            Exception: If the camera returns an empty frame.
        """
        buffer = self._camera.fetch()
        frame = buffer.payload.components[0].data.reshape(self.data_shape)
        if frame.size == 0:
            raise Exception("Camera returned an empty frame")
        data = frame.copy()
        buffer.queue()  # give back buffer to the camera driver
        return data

    @property
    def duration(self) -> Quantity[u.ms]:
        """The duration between the trigger and the end of the exposure.

        Returns ∞ · ms if hardware triggering is used."""
        # TODO: implement hardware triggering.
        return self.exposure.to(u.ms)

    @property
    def exposure(self) -> u.Quantity[u.ms]:
        """Exposure time of the camera"""
        return self._nodes.ExposureTime.value * u.us

    @exposure.setter
    def exposure(self, value: Quantity[u.ms]):
        with self.paused():
            self._nodes.ExposureTime.value = int(value.to_value(u.us))

    @property
    def binning(self) -> int:
        """Pixel binning factor.

        Note:
            setting horizontal and vertical binning separately is not supported.
        """
        return self._nodes.BinningHorizontal.value

    @binning.setter
    def binning(self, value):
        with self.paused():
            if value != self._nodes.BinningHorizontal.value:
                self._nodes.BinningHorizontal.value = int(value)
            if value != self._nodes.BinningVertical.value:
                self._nodes.BinningVertical.value = int(value)

    @property
    def top(self) -> int:
        """The vertical start position of the region of interest (in pixels).

        Note:
            The camera may round up this value to multiples of some power of 2.
        """
        return self._nodes.OffsetY.value

    @top.setter
    def top(self, value: int):
        self._set_round_up(self._nodes.OffsetX, value)

    @property
    def left(self) -> int:
        """The horizontal start position of the region of interest (in pixels).

        Note:
            The camera may round up this value to multiples of some power of 2.
        """
        return self._nodes.OffsetX.value

    @left.setter
    def left(self, value: int):
        self._set_round_up(self._nodes.OffsetX, value)

    def _set_round_up(self, node, value):
        """Sets the value of a property, rounding up to the next multiple of the increment.

        Args:
            node: The camera node to set the value for.
            value: The desired value, which will be rounded up to the next multiple of the node's increment.
        """
        inc = node.inc
        with self.paused():
            # round up value to the next multiple of inc
            node.value = int(value) + ((-value) % inc)

    @property
    def width(self) -> int:
        """Width of the camera frame, in pixels.

        Note:
            The camera may round up this value to multiples of some power of 2.
        """
        return self._nodes.Width.value

    @width.setter
    def width(self, value: int):
        self._set_round_up(self._nodes.Width, value)

    @property
    def height(self) -> int:
        """Height of the camera frame, in pixels.

        Note:
            The camera may round up this value to multiples of some power of 2.
        """
        return self._nodes.Height.value

    @height.setter
    def height(self, value: int):
        self._set_round_up(self._nodes.Height, value)

    @property
    def pixel_size(self) -> Optional[Quantity[u.um]]:
        """Physical pixel size of the camera sensor."""
        return self._pixel_size

    @property
    def data_shape(self):
        """Shape (height, width) of the data array returned by the camera."""
        return self.height, self.width

    @staticmethod
    def enumerate_cameras(cti_file: str):
        """Enumerates all cameras available through the specified GenTL producer.

        Args:
            cti_file: The path to the GenTL producer file.

        Returns:
            list: A list of device information dictionaries for all available cameras.

        Warns:
            UserWarning: If the CTI file cannot be loaded.
        """
        with Harvester() as harvester:
            try:
                harvester.add_file(cti_file, check_validity=True)
                harvester.update()
            except (OSError, FileNotFoundError):
                warnings.warn(f"Failed to load CTI file: {cti_file}")
            return harvester.device_info_list.copy()


class _CameraPause:
    """Context manager for pausing the camera."""

    def __init__(self, camera):
        self._camera = camera

    def __enter__(self):
        return self._camera.stop()

    def __exit__(self, _type, _value, _traceback):
        self._camera.start()



"""A wrapper for the setuptools build backend.

This wrapper adds some custom build steps:
- Extract the version number from pyproject.toml and store it in _version.py
- Extract the git revision from the repository and store it in _version.py
- Update README.md file by compiling it from the docs/source/readme.rst template (only if the source file has changed)
"""

import setuptools.build_meta as build_meta
import subprocess, pathlib


def pre_build(root_path):
    # build README.md if its source file (docs/source/readme.rst) has changed
    try:
        readme_modified = subprocess.check_output(
            ["git", "diff", "source/readme.rst"], text=True, cwd=root_path / "docs"
        ).strip()
        if readme_modified:
            try:
                print("building README.md")
                subprocess.check_call([root_path / "docs" / "make.bat", "markdown"], cwd=root_path / "docs")
            except Exception as e:
                print(f"WARNING: failed to build README.md: {e}")
    except Exception as e:
        print(f"WARNING: failed to run git diff: {e}")

    # get git revision
    try:
        git_revision = subprocess.check_output(["git", "describe", "--tags"], text=True, cwd=root_path).strip()
    except Exception as e:
        git_revision = f"unknown ({e})"

    # get version from pyproject.toml
    try:
        version = "unknown (not specified in pyproject.toml)"
        with open(root_path / "pyproject.toml") as f:
            for line in f.readlines():
                if line.startswith("version"):
                    version = line.split("=")[1].strip().strip('"')
                    break
    except Exception as e:
        version = f"unknown ({e})"

    # write version file
    with open(root_path / "openwfs" / "_version.py", "w") as f:
        f.write(
            "# This file is generated automatically by the build system (uv build, poetry build, see build_backend.py for details)\n"
        )
        f.write(f'__version__ = "{version}"\n')
        f.write('"""Version number reported in the package metadata."""\n\n')
        f.write(f'__git_revision__ = "{git_revision}"\n')
        f.write(
            '"""Git-reported information: Latest tag found in the code branch - number of commits since that tag - hash of current commit."""\n'
        )


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    root_path = pathlib.Path(wheel_directory).parent
    pre_build(root_path)
    return build_meta.build_wheel(wheel_directory, config_settings, metadata_directory)


def build_sdist(sdist_directory, config_settings=None):
    return build_meta.build_sdist(sdist_directory, config_settings)



import time
from collections import deque
from typing import Sequence, Optional, Union

import astropy.units as u
import numpy as np
from astropy.units import Quantity
from numpy.typing import ArrayLike

from ..core import Detector, Processor, PhaseSLM, Actuator
from ..utilities import project, unitless


class PhaseToField(Processor):
    """Takes a phase as input and returns a field

    Computes `amplitude * (exp(1j * phase) + non_modulated_field_fraction)`
    """

    def __init__(
        self,
        slm_phases: Detector,
        field_amplitude: ArrayLike = 1.0,
        non_modulated_field_fraction: float = 0.0,
    ):
        """
        Args:
            slm_phases: The `Detector` that returns the phases of the slm pixels.
            field_amplitude: Field amplitude of the modulated pixels.
            non_modulated_field_fraction: Non-modulated field (e.g. a front reflection).
        """
        super().__init__(slm_phases, multi_threaded=False)
        self.modulated_field_amplitude = field_amplitude
        self.non_modulated_field = non_modulated_field_fraction

    def _fetch(self, slm_phases: np.ndarray) -> np.ndarray:  # noqa
        """
        Updates the complex field output of the SLM. The output field is the sum of the modulated field and the
        non-modulated field.
        """
        return self.modulated_field_amplitude * (np.exp(1j * slm_phases) + self.non_modulated_field)


class _SLMTiming(Detector):
    """Class to simulate the timing of an SLM.

    This class simulates latency (`update_latency`) and
    stabilization (`update_duration`) of the SLM. It does not simulate
    the refresh rate, or the conversion of gray values to phases.
    """

    def __init__(
        self,
        shape: tuple[int, ...],
        update_latency: Quantity[u.ms] = 0.0 * u.ms,
        update_duration: Quantity[u.ms] = 0.0 * u.ms,
    ):

        if len(shape) != 2:
            raise ValueError("Shape of the SLM should be 2-dimensional.")
        super().__init__(
            data_shape=shape,
            pixel_size=Quantity(2.0 / np.min(shape)),
            latency=0 * u.ms,
            duration=0 * u.ms,
            multi_threaded=False,
        )
        self.update_latency = update_latency
        self.update_duration = update_duration

        # _set_point are the voltages that the SLM hardware is currently sending to the display.
        #   The display may not have stabilized yet
        # _state are the voltages currently on the display. They will asymptotically approach _set_point
        # _state_timestamp is the time when the state was last updated
        # _queue is the queue of upcoming frames (set points and their timestamps)
        # that are sent from the PC to the SLM, but not yet displayed.
        self._set_point = np.zeros(shape, dtype=np.float32)
        self._state = np.zeros(shape, dtype=np.float32)
        self._state_timestamp = 0.0
        self._queue = deque()  # Queue to hold phase images and their timestamps

    def _fetch(self):
        return self._update(None)

    def _update(self, append):
        """Computes the currently displayed voltage image, based on the queue of set points and their timestamps.

        This function takes the frames from the queue. The ones that have a time stamp corresponding to
        the current time or earlier are merged into the current _state. The rest is kept.
        Note: if ever converting this object to multi-threading, this function should be protected by a lock.
        """
        current_time = time.time_ns() * u.ns
        while True:
            # peek the first element in the queue
            if not self._queue:
                break
            set_point, timestamp = self._queue[0]
            if timestamp > current_time:
                break

            # we found a frame that is or has been displayed
            # we step the simulation forward to the time of the frame
            self._step_to_time(timestamp)
            self._set_point = set_point
            self._queue.popleft()

        # finally, compute the current state
        if append is None:
            self._step_to_time(current_time)
        else:
            self._queue.append(append)

        return self._state

    def _step_to_time(self, timestamp: Quantity[u.ns]):
        """Step the simulation forward to a given time."""
        if self.update_duration > 0:
            a = np.exp(-(timestamp - self._state_timestamp) / self.update_duration)
            self._state = a * self._state + (1 - a) * self._set_point
        else:
            self._state = self._set_point
        self._state_timestamp = timestamp

    def send(self, phase_image):
        """Send a phase image to the SLM. This method is called by the SLM object.

        The image is not displayed directly. Instead, it is added to a queue with a timestamp
        attached, and it only takes effect after update_latency seconds have passed.

        All old images in the queue are merged with the new image, using an exponential decay.
        """
        display_time = time.time_ns() * u.ns + self.update_latency
        self._update((phase_image, display_time))


class _SLMPhaseResponse(Processor):
    def __init__(self, source: _SLMTiming, phase_response: Optional[np.ndarray] = None):
        super().__init__(source, multi_threaded=False)
        self.phase_response = phase_response

    def _fetch(self, grey_values: np.ndarray) -> np.ndarray:  # noqa
        if self.phase_response is None:
            return grey_values * (2 * np.pi / 256)
        else:
            return self.phase_response[np.rint(grey_values).astype(np.uint8)]


class SLM(PhaseSLM, Actuator):
    """
    A mock version of a phase-only spatial light modulator. Some properties are available to simulate physical
    phenomena such as imperfect phase response, and front reflections (which cause non-modulated light).
    """

    __slots__ = (
        "_hardware_fields",
        "_hardware_phases",
        "_hardware_timing",
        "_back_buffer",
        "refresh_rate",
        "_first_update_ns",
        "_lookup_table",
    )

    def __init__(
        self,
        shape: tuple[int, ...],
        latency: Quantity[u.ms] = 0.0 * u.ms,
        duration: Quantity[u.ms] = 0.0 * u.ms,
        update_latency: Quantity[u.ms] = 0.0 * u.ms,
        update_duration: Quantity[u.ms] = 0.0 * u.ms,
        refresh_rate: Quantity[u.Hz] = 0 * u.Hz,
        field_amplitude: Union[np.ndarray, float, None] = 1.0,
        non_modulated_field_fraction: float = 0.0,
        phase_response: Optional[np.ndarray] = None,
    ):
        """

        Args:
            shape: The 2D shape of the SLM.
            field_amplitude: Field amplitude of the modulated light.
            non_modulated_field_fraction (float): fraction of the field that is not modulated,
                typically due to reflection at the front surface of the SLM. This fraction of
                of the input field is added to the modulated field, i.e. the returned field is
                amplitude*exp(1j*phase) + non_modulated_field_fraction.
            shape: The shape (height, width) of the SLM in pixels
            latency: The latency that the OpenWFS framework uses for synchronization.
            duration: The duration that the OpenWFS framework uses for synchronization.
            update_latency: The latency of the simulated SLM.
                Choose a value different from `latency` to simulate incorrect timing.
            update_duration: The duration of the simulated SLM.
                Choose a value different from `duration` to simulate incorrect timing.
            refresh_rate: Simulated refresh rate. Affects the timing of the `update` method,
                since this will wait until the next vertical retrace. Keep at 0 to disable this feature.
        """
        super().__init__(latency=latency, duration=duration)
        self.refresh_rate = refresh_rate
        # Simulates transferring frames to the SLM
        self._hardware_timing = _SLMTiming(shape, update_latency, update_duration)
        self._hardware_phases = _SLMPhaseResponse(
            self._hardware_timing, phase_response
        )  # Simulates reading the phase from the SLM
        self._hardware_fields = PhaseToField(
            self._hardware_phases, field_amplitude, non_modulated_field_fraction
        )  # Simulates reading the field from the SLM
        self._lookup_table = None  # index = input phase (scaled to -> [0, 255]), value = grey value
        self._first_update_ns = time.time_ns()
        self._back_buffer = np.zeros(shape, dtype=np.float32)

    def update(self):
        """Sends the current phase image to the simulated SLM hardware."""

        self._start()  # wait for detectors to finish
        if self.refresh_rate > 0:
            # wait for the vertical retrace
            time_in_frames = unitless((time.time_ns() - self._first_update_ns) * u.ns * self.refresh_rate)
            time_to_next_frame = (np.ceil(time_in_frames) - time_in_frames) / self.refresh_rate
            time.sleep(time_to_next_frame.tovalue(u.s))
            # update the start time (this is also done in the actual SLM)
            self._start()

        # Apply lookup table and compute grey values from intended phases
        # This uses the same conversion as in the shader:
        # first compute tx = phi * / (2 * pi) + 0.5 / 256
        # use the fractional part of tx to index the lookup table, where:
        # - tx =0 maps to the start of the first element
        # - tx=1-δ maps to the end of the last element
        tx = self._back_buffer * (1 / (2 * np.pi)) + (0.5 / 256)
        tx = tx - np.floor(tx)  # fractional part of tx
        if self._lookup_table is None:
            grey_values = (256 * tx).astype(np.uint8)
        else:
            lookup_index = (self._lookup_table.shape[0] * tx).astype(np.uint8)  # index into lookup table
            grey_values = self._lookup_table[lookup_index]

        self._hardware_timing.send(grey_values)

    @property
    def lookup_table(self) -> Sequence[int]:
        """Lookup table that is used to map the wrapped phase range of 0-2pi to gray values
        (represented in a range from 0 to 256). By default, this is just range(256).
        Note that the lookup table need not contain 256 elements.
        A typical scenario is to use something like `slm.lookup_table=range(142)` to map the 0-2pi range
        to only the first 142 gray values of the slm.
        """
        return self._lookup_table

    @lookup_table.setter
    def lookup_table(self, value: Sequence[int]):
        self._lookup_table = np.asarray(value)

    @property
    def phase_response(self) -> Optional[np.ndarray]:
        """Phase as a function of pixel gray value.

        This lookup table mimics the phase response of the SLM hardware.
        When it is omitted, the phase response is assumed to be linear:
        phase = 2 * pi * grey_value / 256
        """
        return self._hardware_phases.phase_response

    @phase_response.setter
    def phase_response(self, value: Sequence[float]):
        self._hardware_phases.phase_response = value

    def set_phases(self, values: ArrayLike, update=True):
        # no docstring, use documentation from base class

        # Copy the phase image to the back buffer, scaling it as necessary
        project(
            np.atleast_2d(values).astype("float32"),
            out=self._back_buffer,
            source_extent=(2.0, 2.0),
            out_extent=(2.0, 2.0),
        )
        if update:
            self.update()

    @property
    def pixels(self) -> Detector:
        """Returns an object to monitor the current state of the SLM.

        If transient behavior is simulated, these are the effective gray values that are currently displayed on the SLM.
        """
        return self._hardware_timing

    @property
    def field(self) -> Detector:
        """Returns an object to monitor the current state of the SLM.

        If transient behavior is simulated, this is the effective field that is currently on the SLM.
        """
        return self._hardware_fields

    @property
    def phases(self) -> Detector:
        """Returns an object to monitor the current state of the SLM.

        If transient behavior is simulated, this is the effective phase that is currently on the SLM.
        """
        return self._hardware_phases

    @property
    def duration(self) -> Quantity[u.ms]:
        return 0.0 * u.ms



from enum import Enum
from typing import Optional, Sequence

import numpy as np
from numpy.typing import ArrayLike


class WFSResult:
    """
    Data structure for holding wavefront shaping results and statistics.

    Attributes:
        t (ndarray): Measured transmission matrix. If multiple targets were used, the first dimension(s) of `t`
            denote the columns of the transmission matrix (`a` indices), and the last dimensions(s) denote the targets,
            i.e., the rows of the transmission matrix (`b` indices).
        axis (int): Number of dimensions used for denoting a single column of the transmission matrix
            (e.g., 2 dimensions representing the x and y coordinates of the SLM pixels).
        fidelity_noise (ndarray): The estimated loss in fidelity caused by the limited SNR (for each target).
        fidelity_amplitude (ndarray): Estimated reduction of the fidelity due to phase-only modulation (for each target)
            (≈ π/4 for fully developed speckle).
        fidelity_calibration (ndarray): Estimated deviation from a sinusoid response.
        n (int): Total number of segments used in the optimization. When missing, this value is set to the number of
            elements in the first `axis` dimensions of `t`.
        estimated_optimized_intensity (ndarray): When missing, estimated intensity in the target(s) after displaying the
            wavefront correction on a perfect phase-only SLM.
        intensity_offset (Optional[ndarray]): Offset in the signal strength, as a scalar, or as one value per target.
            This is the offset that is caused by a bias in the detector signal, stray light, etc. Default value: 0.0.
    """

    def __init__(
        self,
        t: np.ndarray,
        axis: int,
        fidelity_noise: ArrayLike = 1.0,
        fidelity_amplitude: ArrayLike = 1.0,
        fidelity_calibration: ArrayLike = 1.0,
        n: Optional[int] = None,
        intensity_offset: Optional[ArrayLike] = 0.0,
    ):
        """
        Args:
            t(ndarray): measured transmission matrix.
            axis(int):
                number of dimensions used for denoting a single columns of the transmission matrix
                (e.g. 2 dimensions representing the x and y coordinates of the SLM pixels)
            fidelity_noise(ArrayLike):
                the estimated loss in fidelity caused by the the limited snr (for each target).
            fidelity_amplitude(ArrayLike):
                estimated reduction of the fidelity due to phase-only modulation (for each target)
                (≈ π/4 for fully developed speckle)
            fidelity_calibration(ArrayLike):
                estimated deviation from a sinusoid responds (TODO: experimental, untested)
            n(Optional[int]): total number of segments used in the optimization.
                when missing, this value is set to the number of elements in the first `axis` dimensions of `t`.
            intensity_offset(Optional[ArrayLike]):
                offset in the signal strength, as a scalar, or as one value per target
                this is the offset that is caused by a bias in the detector signal, stray light, etc.
                default value: 0.0

        """
        self.t = t
        self.axis = axis
        self.fidelity_noise = np.atleast_1d(fidelity_noise)
        self.n = np.prod(t.shape[0:axis]) if n is None else n
        self.fidelity_amplitude = np.atleast_1d(fidelity_amplitude)
        self.fidelity_calibration = np.atleast_1d(fidelity_calibration)
        self.estimated_enhancement = np.atleast_1d(
            1.0 + (self.n - 1) * self.fidelity_amplitude * self.fidelity_noise * self.fidelity_calibration
        )
        self.intensity_offset = (
            intensity_offset * np.ones(self.fidelity_calibration.shape)
            if np.isscalar(intensity_offset)
            else intensity_offset
        )
        after = (
            np.sum(np.abs(t), tuple(range(self.axis))) ** 2 * self.fidelity_noise * self.fidelity_calibration
            + intensity_offset
        )
        self.estimated_optimized_intensity = np.atleast_1d(after)

    def __str__(self) -> str:
        noise_warning = "OK" if self.fidelity_noise > 0.5 else "WARNING low signal quality."
        amplitude_warning = "OK" if self.fidelity_amplitude > 0.5 else "WARNING uneven contribution of optical modes."
        calibration_fidelity_warning = (
            "OK" if self.fidelity_calibration > 0.5 else "WARNING non-linear phase response, check " "lookup table."
        )
        return f"""
        Wavefront shaping results:
            fidelity_noise: {self.fidelity_noise} {noise_warning}
            fidelity_amplitude: {self.fidelity_amplitude} {amplitude_warning}
            fidelity_calibration: {self.fidelity_calibration} {calibration_fidelity_warning}
            estimated_enhancement: {self.estimated_enhancement}
            estimated_optimized_intensity: {self.estimated_optimized_intensity}
            """

    def select_target(self, b) -> "WFSResult":
        """
        Returns the wavefront shaping results for a single target

        Args:
            b(int): target to select, as integer index.
                If the target array is multidimensional, it is flattened before selecting the `b`-th component.

        Returns: WFSResults data for the specified target
        """
        return WFSResult(
            t=self.t.reshape((*self.t.shape[0:2], -1))[:, :, b],
            axis=self.axis,
            intensity_offset=self.intensity_offset[:][b],
            fidelity_noise=self.fidelity_noise[:][b],
            fidelity_amplitude=self.fidelity_amplitude[:][b],
            fidelity_calibration=self.fidelity_calibration[:][b],
            n=self.n,
        )

    @staticmethod
    def combine(results: Sequence["WFSResult"]):
        """Merges the results for several sub-experiments.

        Currently, this just computes the average of the fidelities, weighted
        by the number of segments used in each sub-experiment.

        Note: the matrix t is also averaged, but this is not always meaningful.
        The caller can replace the `.t` attribute of the result with a more meaningful value.
        """
        n = sum(r.n for r in results)
        axis = results[0].axis
        if any(r.axis != axis for r in results):
            raise ValueError("All results must have the same axis")

        def weighted_average(attribute):
            data = getattr(results[0], attribute) * results[0].n / n
            for r in results[1:]:
                data += getattr(r, attribute) * r.n / n
            return data

        return WFSResult(
            t=weighted_average("t"),
            n=n,
            axis=axis,
            fidelity_noise=weighted_average("fidelity_noise"),
            fidelity_amplitude=weighted_average("fidelity_amplitude"),
            fidelity_calibration=weighted_average("fidelity_calibration"),
        )

    @property
    def snr(self):
        return 1.0 / (1.0 / self.fidelity_noise - 1.0)


def analyze_phase_stepping(measurements: np.ndarray, axis: int):
    """Analyzes the result of phase stepping measurements, returning matrix `t` and noise statistics

    This function assumes that all measurements were made using the same reference field `A`
    and that the phase of the modulated segment/mode is phase-stepped in equally spaced steps
    between 0 and 2π, whereas the phase of the reference field is kept constant.

    Args:
        measurements(ndarray): array of phase stepping measurements.
            The array holds measured intensities
            with the first one or more dimensions (a1, a2, ...) corresponding to the segments of the SLM,
            one dimension corresponding to the phase steps,
            and the last zero or more dimensions (b1, b2, ...) corresponding to the individual targets
            where the feedback was measured.
        axis(int): indicates which axis holds the phase steps.

    With `phase_steps` phase steps, the measurements are given by

    .. math::

        I_p = \\lvert A + B \\exp(i 2\\pi p / phase_{steps})\\rvert^2,

    This function computes the Fourier transform.

    .. math::

        \\frac{1}{phase_{steps}} \\sum I_p  \\exp(-i 2\\pi p / phase_{steps}) = A^* B

    Returns:
        WFSResult: The result of the analysis. The attribute `t` holds the complex transmission matrix.
            Note that the dimensions of t are reversed with respect to the input, so t has shape b1×b2×...×a1×a2×...
            Other attributes hold fidelity estimates (see WFSResult).
    """
    phase_steps = measurements.shape[axis]
    a_count = int(np.prod(measurements.shape[:axis]))
    a_axes = tuple(range(axis))

    # Fourier transform the phase stepping measurements
    t_f = np.fft.fft(measurements, axis=axis) / phase_steps
    t = np.take(t_f, 1, axis=axis)

    # compute the effect of amplitude variations.
    # for perfectly developed speckle, and homogeneous illumination, this factor will be pi/4
    fidelity_amplitude = np.mean(np.abs(t), a_axes) ** 2 / np.mean(np.abs(t) ** 2, a_axes)

    # estimate the calibration error
    # we first construct a matrix that can be used to fit
    # parameters a and b such that a·t(:) + b·t^*(:) ≈ t_f(:, k, :)
    ff = np.vstack((t.ravel(), np.conj(t.ravel()))).T
    ff_inv = np.linalg.pinv(ff)
    c = np.zeros(phase_steps, np.complex128)

    signal_energy = 0
    for k in range(1, phase_steps):
        cc = ff_inv @ np.take(t_f, k, axis=axis).ravel()
        signal_energy = signal_energy + np.sum(np.abs(ff @ cc) ** 2)
        c[k] = cc[0]
    fidelity_calibration = np.abs(c[1]) ** 2 / np.sum(np.abs(c[1:]) ** 2)

    # TODO: use the pinv fit to estimate the signal strength (requires special treatment of offset)

    # Estimate the error due to noise
    # The signal consists of the response with incorrect modulation,
    # (which occurs twice, ideally in the +1 and -1 components of the Fourier transform),
    # but this factor of two is already included in the 'signal_energy' calculation.
    # an offset, and the rest is noise.
    # average over all targets to get the most accurate result (assuming all targets are similar)
    axes = tuple([i for i in range(t_f.ndim) if i != axis])
    energies = np.sum(np.abs(t_f) ** 2, axis=axes)
    total_energy = np.sum(energies)
    offset_energy = energies[0]
    signal_energy = energies[1] + energies[-1]
    if phase_steps > 3:
        # estimate the noise energy as the energy that is not explained
        # by the signal or the offset.
        noise_energy = (total_energy - signal_energy - offset_energy) / (phase_steps - 3)
        fidelity_noise = np.abs(np.maximum(signal_energy - 2 * noise_energy, 0.0) / signal_energy)
    else:
        fidelity_noise = 1.0  # cannot estimate reliably

    # convert from t_ab to t_ba form
    a_destination = np.array(a_axes) + t.ndim - len(a_axes)
    t = np.moveaxis(t, a_axes, a_destination)

    return WFSResult(
        t,
        axis=axis,
        fidelity_amplitude=fidelity_amplitude,
        fidelity_noise=fidelity_noise,
        fidelity_calibration=fidelity_calibration,
        n=a_count,
    )


class WFSController:
    """
    EXPERIMENTAL - Controller for Wavefront Shaping (WFS) operations using a specified algorithm in the Micro-Manager environment.

    Usage:

    .. code-block:: python

        # not wrapped:
        alg = FourierDualReference(feedback, slm)

        # wrapped
        alg = WFSController(FourierDualReference, feedback, slm)

    Under the hood, a dynamic class is created that inherits both ``WFSController`` and ``FourierDualReference)``.
    Effectively this is similar to having ``class WFSController(FourierDualReference)`` inheritance.

    Since Micro-Manager / PyDevice does not yet support buttons to activate actions, a WFS experiment is started by setting
    the trigger attribute :attr:`wavefront` to the value State.OPTIMIZED
    It adds attributes for inspecting the statistics of the last WFS optimization.
    Manages the state of the wavefront and executes the algorithm to optimize and apply wavefront corrections, while
    exposing all these parameters to MicroManager.
    """

    class State(Enum):
        FLAT = 0
        OPTIMIZED = 1
        REOPTIMIZE = 2

    def __init__(self, _algorithm_class, *args, **kwargs):
        """
        Args:
            algorithm: An instance of a wavefront shaping algorithm.
        """
        super().__init__(*args, **kwargs)
        self._wavefront = WFSController.State.FLAT
        self._result: Optional[WFSResult] = None
        self._feedback_ratio = 0.0
        self._test_wavefront = False  # Trigger to test the optimized wavefront

    def __new__(cls, algorithm_class, *args, **kwargs):
        """Dynamically creates a class of type `class X(WFSController, algorithm_class` and returns an instance of that class"""

        # Dynamically create the class using type()
        class_name = "WFSController_" + algorithm_class.__name__
        DynamicClass = type(class_name, (cls, algorithm_class), {})
        instance = super(WFSController, cls).__new__(DynamicClass)
        return instance

    @property
    def wavefront(self) -> State:
        """
        Enables switching between FLAT or OPTIMIZED wavefront on the SLM.
        Setting this state to OPTIMIZED causes the algorithm execute if the optimized wavefront is not yet computed.
        Setting this state to REOPTIMIZE always causes the algorithm to recompute the wavefront. The state switches to OPTIMIZED after executioin of the algorithm.
        For multi-target optimizations, OPTIMIZED shows the wavefront for the first target.
        """
        return self._wavefront

    @wavefront.setter
    def wavefront(self, value):
        self._wavefront = WFSController.State(value)
        if value == WFSController.State.FLAT:
            self.slm.set_phases(0.0)
        elif value == WFSController.State.OPTIMIZED:
            if self._result is None:
                # run the algorithm
                self._result = self.execute().select_target(0)
            self.slm.set_phases(self.optimized_wavefront)
        else:  # value == WFSController.State.REOPTIMIZE:
            self._result = None  # remove stored result
            self.wavefront = WFSController.State.OPTIMIZED  # recompute the wavefront

    @property
    def fidelity_noise(self) -> float:
        """
        Returns:
            float: the estimated loss in fidelity caused by the limited snr.
        """
        return self._result.fidelity_noise if self._result is not None else 0.0

    @property
    def fidelity_amplitude(self) -> float:
        """
        Returns:
            float: estimated reduction of the fidelity due to phase-only
            modulation (≈ π/4 for fully developed speckle)
        """
        return self._result.fidelity_amplitude if self._result is not None else 0.0

    @property
    def estimated_enhancement(self) -> float:
        """
        Returns:
            float: estimated enhancement: estimated ratio <after>/<before>  (with <> denoting
            ensemble average)
        """
        return self._result.estimated_enhancement if self._result is not None else 0.0

    @property
    def fidelity_calibration(self) -> float:
        """
        Returns:
            float: non-linearity.
        """
        return self._result.fidelity_calibration if self._result is not None else 0.0

    @property
    def estimated_optimized_intensity(self) -> float:
        """
        Returns:
            float: estimated optimized intensity.
        """
        return self._result.estimated_optimized_intensity.mean() if self._result is not None else 0.0

    @property
    def snr(self) -> float:
        """
        Returns:
            float: The average signal-to-noise ratio (SNR) of the wavefront optimization measurements.
        """
        return self._result.snr if self._result is not None else 0.0

    @property
    def optimized_wavefront(self) -> np.ndarray:
        return -np.angle(self._result.t) if self._result is not None else 0.0

    @property
    def feedback_ratio(self) -> float:
        """The ratio of average feedback signals after and before optimization.

        This value is calculated when the :attr:`test_wavefront` trigger is set to True.

        Note: this is *not* the enhancement factor, because the 'before' signal is not ensemble averaged.
            Therefore, this value should be used with caution.

        Returns:
            float: average enhancement of the feedback, 0.0 none if no such enhancement was measured."""
        return self._feedback_ratio

    @property
    def test_wavefront(self) -> bool:
        """Trigger to test the wavefront.

        Set this value `True` to measure feedback signals with a flat and an optimized wavefront and compute the :attr:`feedback_ratio`.
        This value is reset to `False` after the test is performed.
        """
        return False

    @test_wavefront.setter
    def test_wavefront(self, value):
        if value:
            self.wavefront = WFSController.State.FLAT
            feedback_flat = self.feedback.read().sum()
            self.wavefront = WFSController.State.OPTIMIZED
            feedback_shaped = self.feedback.read().sum()
            self._feedback_ratio = float(feedback_shaped / feedback_flat)


class DummyProgressBar:
    """Placeholder for a progress bar object.

    Some functions take an optional tdqm-style progress bar as input.
    This class serves as a placeholder iif no progress bar is given.
    It does nothing.
    """

    def __init__(self):
        self.count = 0

    def update(self):
        pass



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Annotated, Union

import astropy.units as u
import nidaqmx.system
import numpy as np
from annotated_types import Ge, Le
from astropy.units import Quantity
from nidaqmx.constants import TerminalConfiguration, DigitalWidthUnits
from nidaqmx.stream_writers import AnalogMultiChannelWriter

from ..core import Detector
from ..utilities import unitless


@dataclass
class InputChannel:
    """Specification of a NIDAQ input channel for the scanning microscope.

    Attributes:
        channel: The name of the channel, e.g. 'Dev4/ai0'
        v_min: The minimum voltage that can be measured by the channel
        v_max: The maximum voltage that can be measured by the channel
        terminal_configuration: The terminal configuration of the channel,
            defaults to `TerminalConfiguration.DEFAULT`
    """

    channel: str
    v_min: Quantity[u.V]
    v_max: Quantity[u.V]
    terminal_configuration: TerminalConfiguration = TerminalConfiguration.DEFAULT


@dataclass
class Axis:
    """Specification of a NIDAQ output channel for controlling a scan axis.

    Output voltages are clipped to the safety limit indicated by `[v_min, v_max]`
    Note that the actual field of view only covers part of this range (see
    `reference_zoom` in `ScanningMicroscope.`).

    Attributes:
        channel: The name of the channel, e.g. 'Dev4/ao0'
        v_min: The minimum voltage that can safely be sent to the output.
        v_max: The maximum voltage that can safely be sent to the output.
        scale: Conversion factor between voltage at the NI-DAQ card and displacement of the focus in the object plane.
                This may be different for different axes.

        maximum_acceleration: The maximum acceleration of this axis in V per s².
            The output signal will be constructed to ensure that the mirror does not exceed this acceleration,
            except in special cases such as when turning on the system. This acceleration is also used to compute
            how far to overshoot the scan mirror to ensure that it reaches a linear speed over the scan range.
            See `scan` for details.
        terminal_configuration: The terminal configuration of the channel,
            defaults to `TerminalConfiguration.DEFAULT`
    """

    channel: str
    v_min: Quantity[u.V]
    v_max: Quantity[u.V]
    scale: Quantity[u.um / u.V]
    maximum_acceleration: Quantity[u.V / u.s**2]
    terminal_configuration: TerminalConfiguration = TerminalConfiguration.DEFAULT

    def to_volt(self, pos: Union[np.ndarray, float]) -> Quantity[u.V]:
        """Converts relative position [0.0 ... 1.0] to voltage [V_min ... V_max].

        Currently, this is just a linear conversion, but a lookup table may be used in the future.

        Args:
            pos: Relative position value(s) between 0.0 and 1.0.

        Returns:
            Quantity[u.V]: Voltage value(s) between V_min and V_max.
        """
        return self.v_min + np.clip(pos, 0.0, 1.0) * (self.v_max - self.v_min)

    def to_pos(self, volt: Quantity[u.V]) -> np.ndarray:
        """Converts voltage [V_min .. V_max] to relative position [0.0 .. 1.0].

        Args:
            volt: Voltage value(s) to convert.

        Returns:
            np.ndarray: Relative position value(s) between 0.0 and 1.0.
        """
        return unitless((volt - self.v_min) / (self.v_max - self.v_min))

    def maximum_scan_speed(self, linear_range: float) -> Quantity[u.V / u.s]:
        """Computes the maximum scan speed in V per sample.

        It is assumed that the mirror accelerates and decelerates at the maximum
        acceleration, and scans with a constant velocity over the linear range.
        There are two limits to the scan speed:

        - A practical limit: if it takes longer to perform the acceleration + deceleration than
          it does to traverse the linear range, it does not make sense to set the scan speed so high.
          The speed at which acceleration + deceleration takes as long as the linear range is the maximum speed.
        - A hardware limit: when accelerating with the maximum acceleration over a distance
          0.5 · (V_max-V_min) · (1-linear_range),
          the mirror will reach the maximum possible speed.

        Args:
            linear_range: Fraction of the full range that is used for the linear part of the scan.

        Returns:
            Quantity[u.V / u.s]: Maximum scan speed.
        """
        # x = 0.5 · a · t² = 0.5 (v_max - v_min) · (1 - linear_range)
        t_accel = np.sqrt((self.v_max - self.v_min) * (1 - linear_range) / self.maximum_acceleration)
        hardware_limit = t_accel * self.maximum_acceleration

        # t_linear = linear_range · (v_max - v_min) / maximum_speed
        # t_accel = maximum_speed / maximum_acceleration
        # 0.5·t_linear == t_accel => 0.5·linear_range · (v_max-v_min) · maximum_acceleration = maximum_speed²
        practical_limit = np.sqrt(0.5 * linear_range * (self.v_max - self.v_min) * self.maximum_acceleration)
        return np.minimum(hardware_limit, practical_limit)

    def step(self, start: float, stop: float, sample_rate: Quantity[u.Hz]) -> Quantity[u.V]:
        """Generate a voltage sequence to move from `start` to `stop` in the fastest way possible.

        This function assumes that the mirror is standing still at v_start,
        and generates a voltage ramp to move the mirror and stop it exactly at v_end,
        using the maximum acceleration allowed by the mirror.

        The voltage curve is given by:
        v_start + 1/2 a·t²              for t < t_total /2
        v_end - 1/2 a·(t_total-t)²  for t >= t_total

        using continuity at t=t_total/2, we can solve this equation to get:
        t_total = 2·sqrt((v_end - v_start) / a)

        Args:
            start: Starting position in relative coordinates (0.0 to 1.0).
            stop: Ending position in relative coordinates (0.0 to 1.0).
            sample_rate: The sampling rate of the voltage sequence.

        Returns:
            Quantity[u.V]: Voltage sequence for the movement.
        """
        v_start = self.to_volt(start)
        if start == stop:
            return np.zeros((0,)) * u.V  # return empty array
        v_end = self.to_volt(stop)

        # `t` is measured in samples
        # `a` is measured in volt/sample²
        a = self.maximum_acceleration / sample_rate**2 * np.sign(v_end - v_start)
        t_total = unitless(2.0 * np.sqrt((v_end - v_start) / a))
        t = np.arange(np.ceil(t_total + 1e-6))  # add a small number to deal with case t=0 (start=end)
        v_accel = v_start + 0.5 * a * t[: len(t) // 2] ** 2  # acceleration part
        v_decel = v_end - 0.5 * a * (t_total - t[len(t) // 2 :]) ** 2  # deceleration part
        v_decel[-1] = v_end  # fix last point because t may be > t_total due to rounding
        return np.clip(np.concatenate((v_accel, v_decel)), self.v_min, self.v_max)  # noqa ignore incorrect type warning

    def scan(self, start: float, stop: float, sample_count: int, sample_rate: Quantity[u.Hz]) -> tuple:
        """Generate a voltage sequence to scan with a constant velocity from start to stop,
        including acceleration and deceleration.

        Before starting this sequence, the mirror is assumed to be standing still at the launch point,
        which is some distance _before_ start.
        After the scan sequence, the mirror is stopped at the landing point,
        which is some distance _after_ stop.
        The launch point and landing point are returned along with the scan sequence.

        This function also returns a slice object, which represents the part of the sequence
        that corresponds to a linear movement from start to stop. ``slice.stop - slice.start = sample_count``.

        The scan follows the coordinate convention used throughout OpenWFS and Astropy,
        where the coordinates correspond to the centers of the pixels.
        Therefore, while the linear part of the scan starts at start and ends at stop,
        the sample points in this range correspond to the centers of the pixels,
        so the sample at slice.start lies half a pixel _after_ start,
        and the sample at slice.stop - 1 lies half a pixel _before_ stop.

        Args:
            start: Starting position in relative coordinates (0.0 to 1.0).
            stop: Ending position in relative coordinates (0.0 to 1.0).
            sample_count: Number of samples in the linear part of the scan.
            sample_rate: The sampling rate of the voltage sequence.

        Returns:
            tuple: A tuple containing:
                - Quantity[u.V]: Voltage sequence for the entire scan.
                - float: Launch point in relative coordinates.
                - float: Landing point in relative coordinates.
                - slice: Slice object indicating the linear part of the scan.
        """
        v_start = self.to_volt(start)
        if start == stop:  # todo: tolerance?
            return (
                np.ones((sample_count,)) * v_start,
                start,
                start,
                slice(0, sample_count),
            )

        v_end = self.to_volt(stop)
        scan_speed = (v_end - v_start) / sample_count  # V per sample

        # construct a sequence to accelerate from speed 0 to the scan speed
        # we start by constructing a sequence with a maximum acceleration.
        # This sequence may be up to 1  sample longer than needed to reach the scan speed.
        # This last sample is replaced by movement at a linear scan speed
        a = self.maximum_acceleration / sample_rate**2 * np.sign(scan_speed)  # V per sample²
        t_launch = np.arange(np.ceil(unitless(scan_speed / a)))  # in samples
        v_accel = 0.5 * a * t_launch**2  # last sample may have faster scan speed than needed
        if len(v_accel) > 1 and np.abs(v_accel[-1] - v_accel[-2]) > np.abs(scan_speed):
            v_accel[-1] = v_accel[-2] + scan_speed
        v_launch = v_start - v_accel[-1] - 0.5 * scan_speed  # launch point
        v_land = v_end + v_accel[-1] + 0.5 * scan_speed  # landing point

        # linear part of the scan
        v_linear = v_start + scan_speed * (np.arange(sample_count) + 0.5)

        # combine the parts
        v = np.concatenate((v_launch + v_accel, v_linear, v_land - v_accel[::-1]))
        v = np.clip(v, self.v_min, self.v_max)
        launch = self.to_pos(v_launch)
        land = self.to_pos(v_land)
        return v, launch, land, slice(len(v_accel), len(v_accel) + sample_count)

    @staticmethod
    def compute_scale(
        *,
        optical_deflection: Quantity[u.deg / u.V],
        galvo_to_pupil_magnification: float,
        objective_magnification: float,
        reference_tube_lens: Quantity[u.mm],
    ) -> Quantity[u.um / u.V]:
        """Computes the conversion factor between voltage and displacement in the object plane.

        Args:
            optical_deflection: The optical deflection (i.e. twice the mechanical angle) of the mirror
                as a function of applied voltage.
            galvo_to_pupil_magnification: The magnification of the relay system between the galvo mirrors and the pupil.
            objective_magnification: The magnification of the microscope objective.
            reference_tube_lens: The tube lens focal length on which the objective magnification is based.
                This value is manufacturer-specific. Typical values are:
                - 200 mm for Thorlabs, Nikon, Leica, and Mitutoyo
                - 180 mm for Olympus/Evident
                - 165 mm for Zeiss

        Returns:
            Quantity[u.um/u.V]: The conversion factor between voltage and displacement in the object plane.
        """
        f_objective = reference_tube_lens / objective_magnification
        angle_to_displacement = f_objective / u.rad
        return ((optical_deflection / galvo_to_pupil_magnification) * angle_to_displacement).to(u.um / u.V)

    @staticmethod
    def compute_acceleration(
        *,
        optical_deflection: Quantity[u.deg / u.V],
        torque_constant: Quantity[u.N * u.m / u.A],
        rotor_inertia: Quantity[u.kg * u.m**2],
        maximum_current: Quantity[u.A],
    ) -> Quantity[u.V / u.s**2]:
        """Computes the angular acceleration of the focus of the galvo mirror.

        The result is returned in the unit V / second²,
        where the voltage can be converted to displacement using the scale factor.

        Args:
            optical_deflection: The optical deflection (i.e. twice the mechanical angle) of the mirror
                as a function of applied voltage.
            torque_constant: The torque constant of the galvo mirror driving coil.
                May also be given in the equivalent unit of dyne·cm/A.
            rotor_inertia: The moment of inertia of the rotor.
                May also be given in the equivalent unit of g·cm².
            maximum_current: The maximum current that can be applied to the galvo mirror.

        Returns:
            Quantity[u.V / u.s**2]: The maximum acceleration of the galvo mirror in voltage per second squared.
        """
        angular_acceleration = (torque_constant * maximum_current / rotor_inertia).to(u.s**-2) * u.rad
        return (angular_acceleration / optical_deflection).to(u.V / u.s**2)


class TestPatternType(Enum):
    """Type of test pattern to use for simulation."""

    NONE = "none"
    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"
    IMAGE = "image"


class ScanningMicroscope(Detector):
    """Laser scanning microscope with galvo mirrors controlled by a National Instruments data acquisition card (nidaq).

    Effectively, a `ScanningMicroscope` works like a camera, which can be triggered and returns 2-D images.
    These images are obtained by raster-scanning a focus using two galvo mirrors, controlled by a nidaq card,
    and recording a detector signal (typically from a photon multiplier tube (PMT)) with the same card.

    Region of Interest (ROI):
        Upon construction, the maximum voltage range is set for both axes.
        To avoid possible damage to the hardware, this range cannot be exceeded during scanning,
        and the value cannot be changed after construction of the ScanningMicroscope.
        It is recommended to set this voltage range slightly larger than the maximum field of view
        of the microscope (but still within the limits that may cause damage to the hardware),
        so that the mirror has some room to accelerate and decelerate at the edges of the scan range.

        The scan region is defined by the following equation:

        V_start = V_min + (center - 0.5 / (reference_zoom * zoom)) * (V_max - V_min)
        V_stop = V_min + (center + 0.5 / (reference_zoom * zoom)) * (V_max - V_min)

        or, in relative coordinates:

        start_full = center - 0.5 / (reference_zoom * zoom)
        stop_full = center + 0.5 / (reference_zoom * zoom)

        with
        * V_min, V_max: The voltage ranges set for both axes:
        * zoom, reference_zoom: The zoom factors
        * center: The center of the image relative to the full voltage range, default value is 0.5

        The scan region is divided into `resolution × resolution` pixels.
        Within this scan region, a smaller region of interest (ROI) can be defined by setting
        `top`, `left`, `height`, and `width` properties. The ROI is defined in pixels,
        with `(0,0,resolution, resolution)` corresponding to the full field of view.

        start = center + (left / resolution - 0.5) / (reference_zoom * zoom)
        stop = center + ((left + width) / resolution - 0.5) / (reference_zoom * zoom)

    Scan pattern:
        The scanner performs a raster scan, with `y` being the slow axis and `x` the fast axis.
        The pattern is computed such that the mirrors have a constant velocity during the scan.
        The start and end of each scan line, where the mirror accelerates or decelerates are discarded from the data.
        By default, the scanner uses bidirectional scanning along the fast axis,
        which reduces the time needed for a full scan.
        Especially for bidirectional scanning, the synchronization between output and input is crucial, otherwise the
        image will appear teared (even and odd scan lines not aligning). To fine-tune this synchronization, the `delay`
        parameter can be used.
    """

    def __init__(
        self,
        input: InputChannel,
        y_axis: Axis,
        x_axis: Axis,
        sample_rate: Quantity[u.MHz],
        resolution: int,
        reference_zoom: float,
        *,
        delay: Quantity[u.us] = 0.0 * u.us,
        bidirectional: bool = True,
        multi_threaded: bool = True,
        preprocessor: Optional[callable] = None,
        test_pattern: Union[TestPatternType, str] = TestPatternType.NONE,
        test_image=None,
    ):
        """
        Args:
            resolution: number of pixels (height and width) in the full field of view.
                A coarser sampling can be achieved by setting the binning
                Note that the ROI can also be reduced by setting width, height, top and left.
            input: The NI-DAQ channel to use for the input.
            y_axis: The scan axis object for controlling the slow axis.
            x_axis: The scan axis object for controlling the fast axis.
            sample_rate:
                Sample rate of the NI-DAQ input channel.
            delay: Delay between mirror control and data acquisition, measured in microseconds
            reference_zoom: Zoom factor that corresponds to fitting the full field of view exactly.
                The zoom factor in the `zoom` property is multiplied by the `reference_zoom` to compute the scan range.
            bidirectional: If true, enables bidirectional scanning along the fast axis.
            preprocessor: Process the raw data with this function before cropping. When None, the
                preprocessing will be skipped. The function must take input arguments data and sample_rate, and must
                return the preprocessed data.
            test_pattern: Type of test pattern to use for simulation. When set to a value other than 'none', the nidaq hardware is bypassed
                completely, and a test pattern displayed, depending on the value of this parameter:
                - 'horizontal': The voltage that would be sent to the fast axis output channel is used as input value.
                - 'vertical': The voltage that would be sent to the slow axis output channel is used as input value.
                - 'image': The voltages that would be sent are converted to coordinates in an image, resulting in the test image to be returned.
            test_image: The test image to use when `test_pattern` is set to 'image'. This image is expected to be a 2D numpy array
        """
        self._y_axis = y_axis
        self._x_axis = x_axis
        self._input_channel = input
        self._sample_rate = sample_rate.to(u.MHz)
        self._binning = 1  # binning factor
        self._resolution = int(resolution)
        self._roi_top = 0  # in pixels
        self._roi_left = 0  # in pixels
        self._center_x = 0.5  # in relative coordinates (relative to the full field of view)
        self._center_y = 0.5  # in relative coordinates (relative to the full field of view)
        self._delay = delay.to(u.us)
        self._reference_zoom = float(reference_zoom)
        self._zoom = 1.0
        self._bidirectional = bool(bidirectional)
        self._oversampling = 1  # oversampling factor
        self._scan_speed_factor = 0.5  # scan speed relative to maximum

        self._test_pattern = TestPatternType(test_pattern)
        self._test_image = None
        if test_image is not None:
            self._test_image = np.array(test_image, dtype="uint16")
            while self._test_image.ndim > 2:
                self._test_image = np.mean(self._test_image, 2).astype("uint16")

        self._preprocessor = preprocessor

        self._write_task = None
        self._read_task = None

        self._valid = False  # indicates that `trigger()` should initialize the NI-DAQ tasks and scan pattern
        self._scan_pattern = None

        # the pixel size and duration are computed dynamically
        # data_shape just returns self._data shape, and latency = 0.0 ms
        super().__init__(
            data_shape=(resolution, resolution),
            pixel_size=None,
            duration=None,
            latency=0.0 * u.ms,
            multi_threaded=multi_threaded,
        )
        self._update()

    def _update(self):
        """Computes the scan pattern"""

        width = self._data_shape[1]
        height = self._data_shape[0]
        center = 0.5 * self._resolution

        # compute the size of a pixel relative to the maximum voltage range
        actual_zoom = self._reference_zoom * self._zoom
        roi_scale = 1.0 / actual_zoom / self._resolution

        roi_left = self._center_x + (self._roi_left - center) * roi_scale
        roi_right = self._center_x + (self._roi_left + width - center) * roi_scale
        roi_top = self._center_y + (self._roi_top - center) * roi_scale
        roi_bottom = self._center_y + (self._roi_top + height - center) * roi_scale

        # special case for roi width or height of 1.
        # Treat as roi size zero so that the beam does not scan across this single pixel.
        # note: this special handling is not needed for the height, because
        # in the vertical direction the beam does not scan over the pixel,
        # it just steps to the next line at the end of the line
        if width == 1:
            roi_right = 0.5 * (roi_left + roi_right)
            roi_left = roi_right

        # Compute the retrace pattern for the slow axis
        # The scan starts at half a pixel after roi_bottom and ends half a pixel before roi_top
        v_yr = self._y_axis.step(roi_bottom - 0.5 * roi_scale, roi_top + 0.5 * roi_scale, self._sample_rate)

        # Compute the scan pattern for the fast axis
        # The naive speed is the scan speed assuming one pixel per sample
        # The maximum speed is the maximum speed that the mirror can achieve over the scan range
        # (at least, without spending more time on accelerating and decelerating than the scan itself)
        # The user can set the scan speed relative to the maximum speed.
        # If this set speed is lower than naive scan speed, multiple samples are taken per pixel.
        naive_speed = (self._x_axis.v_max - self._x_axis.v_min) * roi_scale * self._sample_rate
        max_speed = self._x_axis.maximum_scan_speed(1.0 / actual_zoom) * self._scan_speed_factor
        if max_speed == 0.0:
            # this may happen if the ROI reaches to or beyond [0,1]. In this case, the mirror has no time to accelerate
            # TODO: implement an auto-adjust option instead of raising an error
            raise ValueError(
                "Maximum scan speed is zero. "
                "This may be because the region of interest exceeds the maximum voltage range"
            )

        self._oversampling = int(np.ceil(unitless(naive_speed / max_speed)))
        oversampled_width = width * self._oversampling
        v_x_even, x_launch, x_land, self._mask = self._x_axis.scan(
            roi_left, roi_right, oversampled_width, self._sample_rate
        )
        if self._bidirectional:
            v_x_odd, _, _, _ = self._x_axis.scan(roi_right, roi_left, oversampled_width, self._sample_rate)
        else:
            v_xr = self._x_axis.step(x_land, x_launch, self._sample_rate)  # horizontal retrace
            v_x_even = np.concatenate((v_x_even, v_xr))
            v_x_odd = v_x_even

        # Set voltages for the scan.
        # The horizontal scan pattern consists of alternating even/odd scan lines
        # For unidirectional mode, these are the same
        # For bidirectional mode, the scan pattern is padded to always have an even number of scan lines
        # The horizontal pattern is repeated continuously, so even during the
        # vertical retrace. In bidirectional scan mode, th
        n_rows = self._data_shape[0] + np.ceil(len(v_yr) / len(v_x_odd)).astype("int32")
        self._n_cols = len(v_x_odd)
        if self._bidirectional and n_rows % 2 == 1:
            n_rows += 1

        scan_pattern = np.zeros((2, n_rows, self._n_cols))
        scan_pattern[1, 0::2, :] = v_x_even  # .reshape((1, -1))
        scan_pattern[1, 1::2, :] = v_x_odd

        y_coord = (np.arange(height) + 0.5) * roi_scale + roi_top
        scan_pattern[0, :height, :] = self._y_axis.to_volt(y_coord).reshape(-1, 1)

        # The last row(s) are used for the vertical retrace
        # We park the mirror after the vertical retrace, to allow
        # the horizontal scan mirror to finish its scan.
        # Note: this may not always be needed, but it guarantees
        # that the horizontal scan mirror is always scanning at the same frequency
        # which is essential for resonant scanning.
        if len(v_yr) > 0:
            retrace = scan_pattern[0, height:, :].reshape(-1)
            retrace[0 : len(v_yr)] = v_yr
            retrace[len(v_yr) :] = v_yr[-1]

        self._scan_pattern = scan_pattern.reshape(2, -1)
        self._valid = True  # indicate that scan patterns have been computed
        if self._test_pattern != TestPatternType.NONE:
            return

        # Sets up NI-DAQ task and i/o channels
        if self._read_task:
            self._read_task.close()
            self._read_task = None
        if self._write_task:
            self._write_task.close()
            self._write_task = None

        self._write_task = ni.Task()
        self._read_task = ni.Task()
        # set the timeout for the nidaq task. Note that this line calls .duration, which
        # in turn calls ._ensure_valid. Therefore, it is important that we gave set self._valid = True
        # above, to avoid an infinite loop.
        self._read_task.in_stream.timeout = self.timeout.to_value(u.s)

        # Configure the sample clock task
        sample_rate = self._sample_rate.to_value(u.Hz)
        sample_count = self._scan_pattern.shape[1]

        # Configure the analog output task (two channels)
        self._write_task.ao_channels.add_ao_voltage_chan(
            self._x_axis.channel,
            min_val=self._x_axis.v_min.to_value(u.V),
            max_val=self._x_axis.v_max.to_value(u.V),
        )
        self._write_task.ao_channels.add_ao_voltage_chan(
            self._y_axis.channel,
            min_val=self._y_axis.v_min.to_value(u.V),
            max_val=self._y_axis.v_max.to_value(u.V),
        )
        self._write_task.timing.cfg_samp_clk_timing(sample_rate, samps_per_chan=sample_count)

        # Configure the analog input task (one channel)
        self._read_task.ai_channels.add_ai_voltage_chan(
            self._input_channel.channel,
            min_val=self._input_channel.v_min.to_value(u.V),
            max_val=self._input_channel.v_max.to_value(u.V),
            terminal_config=self._input_channel.terminal_configuration,
        )
        self._read_task.timing.cfg_samp_clk_timing(sample_rate, samps_per_chan=sample_count)
        self._read_task.triggers.start_trigger.cfg_dig_edge_start_trig(self._write_task.triggers.start_trigger.term)
        delay = self._delay.to_value(u.s)
        if delay > 0.0:
            self._read_task.triggers.start_trigger.delay = delay
            self._read_task.triggers.start_trigger.delay_units = DigitalWidthUnits.SECONDS

        self._writer = AnalogMultiChannelWriter(self._write_task.out_stream)

    def _ensure_valid(self):
        if not self._valid:
            self._update()

    def _do_trigger(self):
        """Makes sure scan patterns are up-to-date, and triggers the NI-DAQ tasks."""
        self._ensure_valid()

        if self._test_pattern != TestPatternType.NONE:
            return

        self._read_task.wait_until_done()
        self._write_task.wait_until_done()

        # write the samples to output in the x-y channels
        self._writer.write_many_sample(self._scan_pattern)

        # Start the tasks
        self._read_task.start()  # waits for trigger coming from the write task
        self._write_task.start()

    def _raw_to_cropped(self, raw: np.ndarray) -> np.ndarray:
        """Converts the raw scanner data back into a 2-dimensional image.

        Because the scanner can return both signed and unsigned integers, both cases are accounted for.
        This function crops the data if padding was added, and it
        flips the even rows back if scanned in bidirectional mode.

        Args:
            raw: The raw 1D array of data from the scanner.

        Returns:
            np.ndarray: A 2D image with the correct dimensions and orientation.

        Raises:
            ValueError: If the data type is not int16 or uint16.
        """
        # convert data to 2-d, discard padding
        cropped = raw.reshape(-1, self._n_cols)[: self._data_shape[0], self._mask]

        # down sample along fast axis if needed
        if self._oversampling > 1:
            # remove samples if not divisible by oversampling factor
            cropped = cropped[:, : (cropped.shape[1] // self._oversampling) * self._oversampling]
            cropped = cropped.reshape(cropped.shape[0], -1, self._oversampling)
            cropped = np.round(np.mean(cropped, 2)).astype(cropped.dtype)  # todo: faster alternative?

        # Change the data type into uint16 if necessary
        if cropped.dtype == np.int16:
            # add 32768 to go from -32768-32767 to 0-65535
            cropped = cropped.view("uint16") + 0x8000
        elif cropped.dtype != np.uint16:
            raise ValueError(f"Only int16 and uint16 data types are supported at the moment, got type {cropped.dtype}.")

        if self._bidirectional:  # note: requires the mask to be symmetrical
            cropped[1::2, :] = cropped[1::2, ::-1]

        return cropped

    def _fetch(self) -> np.ndarray:  # noqa
        """Reads the acquired data from the input task.

        This method retrieves the data from the NI-DAQ input task, processes it according to the
        test pattern settings, and applies any preprocessing function if specified.

        Returns:
            np.ndarray: The processed image data from the scanning microscope.

        Raises:
            ValueError: If an invalid test pattern is specified or if no test image was provided
                when using the 'image' test pattern.
        """
        if self._test_pattern is TestPatternType.NONE:
            raw = self._read_task.in_stream.read()
            self._read_task.stop()
            self._write_task.stop()
        elif self._test_pattern == TestPatternType.HORIZONTAL:
            raw = np.round(self._x_axis.to_pos(self._scan_pattern[1, :] * u.V) * 10000).astype("int16")
        elif self._test_pattern == TestPatternType.VERTICAL:
            raw = np.round(self._y_axis.to_pos(self._scan_pattern[0, :] * u.V) * 10000).astype("int16")
        elif self._test_pattern == TestPatternType.IMAGE:
            if self._test_image is None:
                raise ValueError("No test image was provided for the image simulation.")
            # todo: cache the test image
            row = np.floor(
                self._y_axis.to_pos(self._scan_pattern[0, :] * u.V) * (self._test_image.shape[0] - 1)
            ).astype("int32")
            column = np.floor(
                self._x_axis.to_pos(self._scan_pattern[1, :] * u.V) * (self._test_image.shape[1] - 1)
            ).astype("int32")
            raw = self._test_image[row, column]
        else:
            raise ValueError(
                f"Invalid simulation option {self._test_pattern}. "
                "Should be 'horizontal', 'vertical', 'image', or 'None'"
            )

        # Preprocess raw data if a preprocess function is set
        if self._preprocessor is None:
            preprocessed_raw = raw
        elif callable(self._preprocessor):
            preprocessed_raw = self._preprocessor(data=raw, sample_rate=self._sample_rate)
        else:
            raise TypeError(f"Invalid type for {self._preprocessor}. Should be callable or None.")
        return self._raw_to_cropped(preprocessed_raw)

    def close(self):
        """Close connection to the NI-DAQ.

        This method closes both the read and write tasks, releasing the hardware resources.
        It should be called when the scanning microscope is no longer needed.
        """
        self._read_task.close()
        self._write_task.close()

    @property
    def test_pattern(self) -> TestPatternType:
        """Get the current test pattern type.

        Returns:
            TestPatternType: The current test pattern setting.
        """
        return self._test_pattern

    @test_pattern.setter
    def test_pattern(self, value: TestPatternType):
        """Set the test pattern type.

        Args:
            value: The test pattern to use. Can be one of the TestPatternType enum values
                or a string ('none', 'horizontal', 'vertical', 'image').
        """
        self._test_pattern = TestPatternType(value)

    @property
    def preprocessor(self) -> Optional[callable]:
        """An optional function to preprocess raw data before cropping.

        The function takes a linear array of raw data as required arguments,
        and a list of keyword arguments. Currently, the following arguments are passed:
            - data: The raw data array from the scanner.
            - sample_rate (Quantity[u.MHz]): The sample rate of the NI-DAQ input channel.

        Returns:
            Optional[callable]: The current preprocessor function or None if not set.
        """
        return self._preprocessor

    @preprocessor.setter
    def preprocessor(self, value: Optional[callable]):
        """Set the preprocessor function.

        Args:
            value: A callable function that takes raw data and sample_rate as arguments,
                or None to disable preprocessing.

        Raises:
            TypeError: If the value is not callable and not None.
        """
        if not callable(value) and value is not None:
            raise TypeError(f"Invalid type for {self._preprocessor}. Should be callable or None.")
        self._preprocessor = value

    @property
    def pixel_size(self) -> Quantity:
        """The size of a pixel in the object plane.

        This property calculates the physical size of each pixel based on the voltage range,
        scale factors of the axes, zoom settings, and resolution.

        Returns:
            Quantity: The physical size of each pixel in the object plane, as a 2D quantity (y, x).
        """
        # TODO: make extent a read-only attribute of Axis
        extent_y = (self._y_axis.v_max - self._y_axis.v_min) * self._y_axis.scale
        extent_x = (self._x_axis.v_max - self._x_axis.v_min) * self._x_axis.scale
        return (Quantity(extent_y, extent_x) / (self._reference_zoom * self._zoom * self._resolution)).to(u.um)

    @property
    def duration(self) -> Quantity[u.ms]:
        """Total duration of scanning for one frame.

        This property calculates the time required to complete a full scan based on
        the scan pattern length and the sample rate.

        Returns:
            Quantity[u.ms]: The total time required to complete one frame.
        """
        self._ensure_valid()  # make sure _scan_pattern is up-to-date
        return (self._scan_pattern.shape[1] / self._sample_rate).to(u.ms)

    @property
    def left(self) -> int:
        """The leftmost pixel of the Region of Interest (ROI) in the scan range.

        Returns:
            int: The x-coordinate of the left edge of the ROI in pixels.
        """
        return self._roi_left

    @left.setter
    def left(self, value: int):
        """Set the leftmost pixel of the ROI.

        Args:
            value: The x-coordinate of the left edge of the ROI in pixels.
        """
        self._roi_left = int(value)
        self._valid = False

    @property
    def top(self) -> int:
        """The topmost pixel of the ROI in the scan range.

        Returns:
            int: The y-coordinate of the top edge of the ROI in pixels.
        """
        return self._roi_top

    @top.setter
    def top(self, value: int):
        """Set the topmost pixel of the ROI.

        Args:
            value: The y-coordinate of the top edge of the ROI in pixels.
        """
        self._roi_top = int(value)
        self._valid = False

    @property
    def height(self) -> int:
        """The number of pixels in the vertical dimension of the ROI.

        Returns:
            int: The height of the ROI in pixels.
        """
        return self._data_shape[0]

    @height.setter
    def height(self, value: int):
        """Set the height of the ROI.

        Args:
            value: The height of the ROI in pixels.

        Raises:
            ValueError: If the value is less than 1 or greater than the resolution.
        """
        if value < 1 or value > self._resolution:
            raise ValueError(f"Height must be between 1 and {self._resolution}")
        self._data_shape = (int(value), int(self.data_shape[1]))
        self._valid = False

    @property
    def width(self) -> int:
        """The number of pixels in the horizontal dimension of the ROI.

        Depending on the scan speed and sample rate, the scanner may
        acquire multiple data points along a scan line, and return the
        averaged value.
        A value of 1 is treated as a special case,
        where the beam does not move horizontally at all
        (i.e. it does not scan back and forth over the size of this single pixel).

        Returns:
            int: The width of the ROI in pixels.
        """
        return self.data_shape[1]

    @width.setter
    def width(self, value: int):
        """Set the width of the ROI.

        Args:
            value: The width of the ROI in pixels.

        Raises:
            ValueError: If the value is less than 1 or greater than the resolution.
        """
        if value < 1 or value > self._resolution:
            raise ValueError(f"Width must be between 1 and {self._resolution}")
        self._data_shape = (self.data_shape[0], int(value))
        self._valid = False

    def reset_roi(self):
        """Reset the ROI to span the original left, top, width and height.

        This method resets the Region of Interest to cover the full scan area,
        setting left and top to 0, and width and height to the full resolution.
        """
        self.left = 0
        self.top = 0
        self.width = self._resolution
        self.height = self._resolution

    @property
    def dwell_time(self) -> Quantity[u.us]:
        """The time spent on each pixel during scanning.

        This property calculates the dwell time based on the oversampling factor
        and the sample rate of the scanner.

        Returns:
            Quantity[u.us]: The time spent on each pixel during scanning.
        """
        return (self._oversampling / self._sample_rate).to(u.us)

    @property
    def delay(self) -> Quantity[u.us]:
        """Delay between the control signal to the mirrors and the start of data acquisition.

        This property controls the synchronization between the mirror movement and data acquisition.
        Adjusting this value can help correct for timing mismatches that cause image tearing.

        Returns:
            Quantity[u.us]: The current delay setting.
        """
        return self._delay

    @delay.setter
    def delay(self, value: Quantity[u.us]):
        """Set the delay between mirror control and data acquisition.

        Args:
            value: The delay time to set, in microseconds.
        """
        self._delay = value
        self._valid = False

    @property
    def exposure(self) -> Quantity[u.ms]:
        """The required time to scan a frame.

        This is an alias for the duration property, provided for compatibility with
        the camera interface.

        Returns:
            Quantity[u.ms]: The total time required to complete one frame.
        """
        return self.duration

    @property
    def bidirectional(self) -> bool:
        """Whether scanning is bidirectional along the fast axis.

        When enabled, the scanner acquires data in both directions along the fast axis,
        which reduces the total scan time but may require more precise timing adjustments.

        Returns:
            bool: True if bidirectional scanning is enabled, False otherwise.
        """
        return self._bidirectional

    @bidirectional.setter
    def bidirectional(self, value: bool):
        """Set the bidirectional scanning mode.

        Args:
            value: True to enable bidirectional scanning, False to use unidirectional scanning.
        """
        self._bidirectional = value
        self._valid = False

    @property
    def zoom(self) -> float:
        """Zoom factor for the scanning microscope.

        The zoom factor determines the pixel size relative to the original pixel size.
        When zooming in or out, the center of the region of interest is kept constant.
        Note that this may cause the field of view to get extended to outside the original FOV.

        Returns:
            float: The current zoom factor.
        """
        return self._zoom

    @zoom.setter
    def zoom(self, value: float):
        """Set the zoom factor.

        Args:
            value: The new zoom factor to set. Higher values result in smaller field of view
                and higher resolution.
        """
        # compute how far the roi center is away from the _center_x before the zoom
        roi_scale = 1.0 / (self._reference_zoom * self._zoom) / self._resolution
        center_y_before = (self._roi_top + 0.5 * self._data_shape[0]) * roi_scale
        center_x_before = (self._roi_left + 0.5 * self._data_shape[1]) * roi_scale

        # compute how far it will be from _center_x after adjusting the zoom
        center_y_after = center_y_before * self._zoom / value
        center_x_after = center_x_before * self._zoom / value

        # correct the center position such that the center of the roi does not move
        self._center_y += center_y_before - center_y_after
        self._center_x += center_x_before - center_x_after

        self._zoom = float(value)
        self._valid = False

    @property
    def offset_x(self) -> float:
        """The center of the full field of view in the horizontal direction.

        The offset is relative to the full voltage range specified in the Axis objects,
        with 0.0 corresponding to the center of the voltage range,
        and -0.5 and +0.5 to the edges of the voltage range.

        Note that changing the offset may cause the ROI to move outside the original field of view.
        Also, it may cause the scan speed to change, as the mirror has a shorter distance to accelerate or decelerate.
        """
        return self._center_x - 0.5

    @offset_x.setter
    def offset_x(self, value: float):
        self._center_x = float(value) + 0.5
        self._valid = False

    @property
    def offset_y(self) -> float:
        """The center of the full field of view in the vertical direction.

        The offset is relative to the full voltage range specified in the Axis objects,
        with 0.0 corresponding to the center of the voltage range,
        and -0.5 and +0.5 to the edges of the voltage range.

        Note that changing the offset may cause the ROI to move outside the original field of view.
        Also, it may cause the scan speed to change, as the mirror has a shorter distance to accelerate or decelerate.
        """
        return self._center_y - 0.5

    @offset_y.setter
    def offset_y(self, value: float):
        self._center_y = float(value) + 0.5
        self._valid = False

    @property
    def resolution(self) -> int:
        return self._resolution

    @resolution.setter
    def resolution(self, value: int):
        self._scale_roi(value / self._resolution)

    def _scale_roi(self, factor: float):
        """Adjusts resolution, top, left, width and height by the same factor."""

        def adjust(x):
            return int(np.round(x * factor))

        self._roi_left = adjust(self._roi_left)
        self._roi_top = adjust(self._roi_top)
        self._data_shape = (adjust(self._data_shape[0]), adjust(self._data_shape[1]))
        self._resolution = adjust(self._resolution)
        self._valid = False

    @property
    def binning(self) -> int:
        """Undersampling factor.

        Increasing the binning reduces the number of pixels in the image while keeping dwell time the same.
        As a result, the total duration of a scan decreases.

        Note:
            This behavior is different from that of a real camera.
            No actual binning is performed, the scanner just takes fewer steps in x and y

        Note: the ROI is kept the same as much as possible.
            However, due to rounding, it may vary slightly.

        """
        return self._binning

    @binning.setter
    def binning(self, value: int):
        if value < 1:
            raise ValueError("Binning value should be a positive integer")
        self._scale_roi(self._binning / int(value))
        self._binning = int(value)

    @property
    def scan_speed(self) -> Annotated[float, Ge(0.05), Le(1.0)]:
        """The scan speed relative to the maximum scan speed."""
        return self._scan_speed_factor

    @scan_speed.setter
    def scan_speed(self, value):
        self._scan_speed_factor = np.clip(float(value), 0.05, 1.0)

    @staticmethod
    def list_devices() -> list:
        """Returns a list of all NI-DAQ devices available on the system.

        This method queries the system for all National Instruments data acquisition devices
        that are currently connected and available.

        Returns:
            list: A list of device names (strings) that can be used in the channel specifications.
        """
        return [d.name for d in nidaqmx.system.System().devices]



from concurrent.futures import Future
from typing import Sequence, Union

import numpy as np

from ..core import Detector


class HDRCamera(Detector):
    """Wrapper to convert a camera to a high dynamic range (HDR) camera.

    This wrapper forwards all method calls and attribute acces to the camera object,
    except for the trigger method, which is replaced by a method that triggers the camera
    multiple times with different exposure times.

    todo: find out why the HDR camera is so slow.

    Args:
        camera: The camera object to wrap.
        background: The background value to subtract from each image before scaling.
            If a sequence is passed, each value corresponds to the background for a different exposure factor.
        saturation_threshold: The threshold value to consider a pixel saturated, in which case it will not be taken into account in the computation of the result.
        exposure_factors: A sequence of exposure factors to use for HDR imaging.
    """

    def __init__(
        self,
        camera,
        background: Union[Sequence[float], float],
        saturation_threshold: int,
        exposure_factors: Sequence[float] = (1.0, 0.1, 0.01),
    ):
        self._camera = camera  # object.__setattr__(self, "_camera", camera)
        self._exposure_factors = np.asarray(exposure_factors, dtype=np.float32)
        self._saturation_threshold = saturation_threshold
        self._background = (background,) * len(exposure_factors) if np.isscalar(background) else background
        super().__init__(
            data_shape=camera.data_shape,
            pixel_size=camera.pixel_size,
            duration=camera.duration,
            latency=camera.latency,
            multi_threaded=False,
        )

    def trigger(self, *args, out=None, immediate=False, **kwargs) -> Future:
        """Trigger the camera multiple times with different exposure times.

        todo: at the moment, this method is always blocking, the immediate flag is ignored.

        Args:
            *args: Positional arguments to pass to the camera's trigger method.
            out: Optional output buffer.
            immediate: If True, trigger the camera immediately.
            **kwargs: Keyword arguments to pass to the camera's trigger method.

        Returns:
            A Future object representing the result of the trigger operation.
        """
        exposure = self._camera.exposure
        data = None
        weight = None
        for factor, background in zip(self._exposure_factors, self._background):
            self._camera.exposure = exposure * factor
            frame = (self._camera.trigger(*args, out=out, immediate=True, **kwargs).result()).astype(
                np.float32
            ) - background
            mask = frame < self._saturation_threshold - background
            frame[~mask] = 0
            if data is None:
                data = frame
                weight = mask * factor
            else:
                data += frame
                weight += mask * factor

        # finally divide by zero, special treatment if there were zeros in the mask
        mask = weight == 0
        weight[mask] = 1.0
        data[mask] = self._saturation_threshold
        data /= weight
        if out is not None:
            out[...] = data

        # Set the exposure back to the original value
        self._camera.exposure = exposure
        result = Future()
        result.set_result(data)
        return result

    def _fetch(self, *args, **kwargs) -> np.ndarray:
        """Fetch method required by the Detector interface.

        This method is not implemented for HDRCamera because the trigger method
        directly returns the result without using the fetch mechanism.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            NotImplementedError: Always raised because this method is not implemented.
        """
        raise NotImplementedError("Fetch is not implemented for HDRCamera")

    def __getattr__(self, name):
        """Forward attribute access to the wrapped camera object.

        This method is called when an attribute is not found in the HDRCamera instance.
        It forwards the attribute access to the wrapped camera object, except for
        attributes starting with an underscore.

        Args:
            name: The name of the attribute to access.

        Returns:
            The attribute value from the wrapped camera object.
        """
        # Forward attribute access to the target
        if name.startswith("_"):
            return super().__getattr__(name)
        else:
            return getattr(self._camera, name)

    def __setattr__(self, name, value):
        """Forward attribute setting to the wrapped camera object.

        This method is called when an attribute is set on the HDRCamera instance.
        It forwards the attribute setting to the wrapped camera object, except for
        attributes starting with an underscore.

        Args:
            name: The name of the attribute to set.
            value: The value to set the attribute to.
        """
        # Forward attribute setting to the target
        if name.startswith("_"):
            return super().__setattr__(name, value)
        else:
            setattr(self._camera, name, value)



from typing import Optional

import numpy as np

from .dual_reference import DualReference
from .utilities import analyze_phase_stepping
from ..core import Detector, PhaseSLM
from ..utilities import tilt


class FourierDualReference(DualReference):
    """Fourier double reference algorithm, based on Mastiani et al. [1].

    Improvements over [1]:

    - The set of plane waves is taken from a disk in k-space instead of a square.
    - No overlap between the two halves is needed, instead the final stitching step is done
      using measurements already in the data set.
    - When only a single target is optimized, the algorithm can be used in an iterative version
      to increase SNR during the measurement, similar to [2].

    [1]: Bahareh Mastiani, Gerwin Osnabrugge, and Ivo M. Vellekoop,
    "Wavefront shaping for forward scattering," Opt. Express 30, 37436-37445 (2022)

    [2]: X. Tao, T. Lam, B. Zhu, et al., “Three-dimensional focusing through scattering media using conjugate adaptive
    optics with remote focusing (CAORF),” Opt. Express 25, 10368–10383 (2017).
    """

    def __init__(
        self,
        *,
        feedback: Detector,
        slm: PhaseSLM,
        slm_shape=(500, 500),
        phase_steps=4,
        k_radius: float = 3.2,
        k_step: float = 1.0,
        iterations: int = 2,
        amplitude: np.ndarray = 1.0,
        optimized_reference: Optional[bool] = None,
    ):
        """
        Args:
            feedback: Source of feedback
            slm: The spatial light modulator
            slm_shape: The shape that the SLM patterns & transmission matrices are calculated for,
                does not necessarily have to be the actual pixel dimensions as the SLM.
            phase_steps: The number of phase steps.
            k_radius: Limit grid points to lie within a circle of this radius.
            k_step: Make steps in k-space of this value. 1 corresponds to diffraction limited tilt.
            iterations: Number of ping-pong iterations. Defaults to 2.
            amplitude: Amplitude profile over the SLM. Defaults to 1.0 (flat)
            optimized_reference:
                When `True`, during each iteration the other half of the SLM displays the optimized pattern so far (as in [1]).
                When `False`, the algorithm optimizes A with a flat wavefront on B,
                and then optimizes B with a flat wavefront on A.
                This mode also allows for multi-target optimization.
                When set to `None` (default), the algorithm uses True if there is a single target,
                and False if there are multiple targets.


        """
        self._k_radius = k_radius
        self._k_step = k_step
        self._shape = slm_shape
        group_mask = np.zeros(slm_shape, dtype=bool)
        group_mask[:, slm_shape[1] // 2 :] = True
        super().__init__(
            feedback=feedback,
            slm=slm,
            phase_patterns=self._construct_modes(),
            group_mask=group_mask,
            phase_steps=phase_steps,
            iterations=iterations,
            amplitude=amplitude,
            optimized_reference=optimized_reference,
        )

    def _construct_modes(self) -> tuple[np.ndarray, np.ndarray]:
        """Constructs the set of plane wave modes."""

        # start with a grid of k-values
        # then filter out the ones that are outside the circle
        # in the grid, the spacing in the kx direction is twice the spacing in the ky direction
        # because we subdivide the SLM into two halves along the x direction,
        # which effectively doubles the number of kx values
        int_radius_x = np.ceil(self.k_radius / (self.k_step * 2))
        int_radius_y = np.ceil(self.k_radius / self.k_step)
        kx, ky = np.meshgrid(
            np.arange(-int_radius_x, int_radius_x + 1) * (self.k_step * 2),
            np.arange(-int_radius_y, int_radius_y + 1) * self.k_step,
        )

        # only keep the points within the circle
        mask = kx**2 + ky**2 <= self.k_radius**2
        k = np.stack((ky[mask], kx[mask])).T

        # construct the modes for these kx ky values
        modes = np.zeros((len(k), *self._shape), dtype=np.float32)
        for i, k_i in enumerate(k):
            # tilt generates a pattern from -2.0 to 2.0 (The convention for Zernike modes normalized to an RMS of 1).
            # The natural step to take is the Abbe diffraction limit of the modulated part,
            # which corresponds to a gradient from -π to π over the modulated part.
            # TODO: modify tilt to take a 2-D argument, returning the mode set directly?
            modes[i] = tilt(self._shape, g=k_i * 0.5 * np.pi)

        return modes, modes

    @property
    def k_radius(self) -> float:
        """The maximum radius of the k-space circle."""
        return self._k_radius

    @k_radius.setter
    def k_radius(self, value: float):
        """Sets the maximum radius of the k-space circle, triggers the building of the internal k-space properties."""
        self._k_radius = float(value)
        self.phase_patterns = self._construct_modes()

    @property
    def k_step(self) -> float:
        return self._k_step

    @k_step.setter
    def k_step(self, value: float):
        self._k_step = float(value)
        self.phase_patterns = self._construct_modes()



import time
from typing import Sequence, Optional

import astropy.units as u
import numpy as np
from astropy.units import Quantity

from ..core import Detector, Processor, Actuator
from ..processors import CropProcessor
from ..utilities import ExtentType, get_pixel_size, set_pixel_size, unitless


class StaticSource(Detector):
    """
    Detector that returns pre-set data. Also simulates latency and measurement duration.
    """

    def __init__(
        self,
        data: np.ndarray,
        *,
        pixel_size: Optional[ExtentType] = None,
        extent: Optional[ExtentType] = None,
        latency: Quantity[u.ms] = 0 * u.ms,
        duration: Quantity[u.ms] = 0 * u.ms,
        multi_threaded: bool = None,
    ):
        """
        Initializes the MockSource
        TODO: factor out the latency and duration into a separate class?
        Args:
            data (np.ndarray): The pre-set data to be returned by the mock source.
            pixel_size (Quantity, optional): The size of each pixel in the data.
                If not specified, the pixel size is calculated from the extent and the data shape.
                If neither pixel size nor extent are specified, the pixel size from `data` is used, if available.
                Otherwise, pixel_size is set to None.
            extent (Optional[ExtentType]): The physical extent of the data array.
                Only used when the pixel size is not specified explicitly.
        """
        if pixel_size is None:
            if extent is not None:
                pixel_size = Quantity(extent) / data.shape
            else:
                pixel_size = get_pixel_size(data)
        else:
            data = set_pixel_size(data, pixel_size)  # make sure the data array holds the pixel size

        if pixel_size is not None and (np.isscalar(pixel_size) or pixel_size.size == 1) and data.ndim > 1:
            pixel_size = pixel_size.repeat(data.ndim)

        if multi_threaded is None:
            multi_threaded = latency > 0 * u.ms or duration > 0 * u.ms

        self._data = data
        super().__init__(
            data_shape=data.shape,
            pixel_size=pixel_size,
            latency=latency,
            duration=duration,
            multi_threaded=multi_threaded,
        )

    def _fetch(self) -> np.ndarray:  # noqa
        total_time_s = self.latency.to_value(u.s) + self.duration.to_value(u.s)
        if total_time_s > 0.0:
            time.sleep(total_time_s)
        return self._data

    @property
    def data(self):
        """
        The pre-set data to be returned by the mock source.

        Note:
            The data is not copied.
            When setting the `data` property, the `data_shape` attribute is updated accordingly.
            If the `data` property is set with an array that has the `pixel_size` metadata set,
            the `pixel_size` attribute is also updated accordingly.
        """
        return self._data

    @data.setter
    def data(self, value):
        self._data = value
        self._pixel_size = get_pixel_size(value)
        self._data_shape = value.shape


class NoiseSource(Detector):
    def __init__(
        self,
        noise_type: str,
        *,
        data_shape: tuple[int, ...],
        pixel_size: Quantity,
        multi_threaded=True,
        generator=None,
        **kwargs,
    ):
        self._noise_type = noise_type
        self._noise_arguments = kwargs
        self._rng = generator if generator is not None else np.random.default_rng()
        super().__init__(
            data_shape=data_shape,
            pixel_size=pixel_size,
            latency=0 * u.ms,
            duration=0 * u.ms,
            multi_threaded=multi_threaded,
        )

    def _do_trigger(self) -> None:
        pass  # no hardware triggering is needed for this mock device

    def _fetch(self) -> np.ndarray:  # noqa
        if self._noise_type == "uniform":
            return self._rng.uniform(**self._noise_arguments, size=self.data_shape)
        elif self._noise_type == "gaussian":
            return self._rng.normal(**self._noise_arguments, size=self.data_shape)
        else:
            raise ValueError(f"Unknown noise type: {self._noise_type}")

    @Detector.data_shape.setter
    def data_shape(self, value):
        self._data_shape = tuple(value)


class ADCProcessor(Processor):
    """Mimics an analog-digital converter.

    At the moment, only positive input and output values are supported.
    """

    def __init__(
        self,
        source: Detector,
        analog_max: Optional[float],
        digital_max: int = 0xFFFF,
        amplifier_bias: float = 0.0,
        shot_noise: bool = False,
        gaussian_noise_std: float = 0.0,
        multi_threaded: bool = True,
        generator=None,
    ):
        """
        Initializes the ADCProcessor class, which mimics an analog-digital converter.

        Args:
            source: The source detector providing analog data.
            analog_max: The maximum analog value that can be handled by the ADC.
                If set to None, each measurement will be automatically scaled so that the maximum
                value in the data set returned by `source` is converted to `digital_max`.
                Note that this means that the values from two different measurements cannot be compared quantitatively.

            digital_max:
                The maximum digital value that the ADC can output, default is unsigned 16-bit maximum.

            amplifier_bias:
                The bias value to add to the data before digitization. This bias is applied after scaling
                from analog to digital values, and after adding the noise, but before rounding and cropping.

            shot_noise:
                Flag to determine if Poisson noise should be applied instead of rounding.
                Useful for realistically simulating detectors.

            gaussian_noise_std:
                If >0, add gaussian noise with std of this value to the data.
        """
        super().__init__(source, multi_threaded=multi_threaded)
        self._analog_max = None
        self._digital_max = None
        self._amplifier_bias = amplifier_bias
        self._shot_noise = False
        self._gaussian_noise_std = 0.0
        self._rng = generator if generator is not None else np.random.default_rng()
        self.gaussian_noise_std = gaussian_noise_std
        self.shot_noise = shot_noise
        self.analog_max = analog_max  # check value
        self.digital_max = digital_max  # check value
        self._scale = 1.0  # additional scaling factor for the data, used in the _fetch method

    def _fetch(self, data) -> np.ndarray:  # noqa
        """Clips the data to the range of the ADC, and digitizes the values."""

        if self.analog_max is None:  # auto scaling
            max_value = np.max(data)
            if max_value > 0.0:
                data = data * (self.digital_max / max_value * self._scale)  # auto-scale to maximum value
        else:
            data = data * (self.digital_max / self.analog_max * self._scale)

        if self._shot_noise:
            data = self._rng.poisson(data)

        if self._gaussian_noise_std > 0.0:
            data = data + self._rng.normal(scale=self._gaussian_noise_std, size=data.shape)

        if self._amplifier_bias != 0.0:
            data = data + self._amplifier_bias

        return np.clip(np.rint(data), 0, self.digital_max).astype("uint16")

    @property
    def analog_max(self) -> Optional[float]:
        """Maximum value that the ADC can handle as input

        This value and all higher values are converted to `digital_max`.
        When set to 0.0, the input signal is scaled automatically so that the maximum corresponds to
        `digital_max`
        """
        return self._analog_max

    @analog_max.setter
    def analog_max(self, value: Optional[float]):
        if value is None:
            self._analog_max = None
            return
        if value < 0.0:
            raise ValueError("analog_max cannot be negative")
        self._analog_max = float(value)

    @property
    def digital_max(self) -> int:
        """Maximum value that the ADC can output.

        Default value is 0xFFFF (16 bits)
        """
        return self._digital_max

    @property
    def conversion_factor(self) -> Optional[float]:
        """Conversion factor between analog and digital values.
        If analog_max is set to None, each frame is auto-scaled, and this function returns None.
        """
        return self.digital_max / self.analog_max if self.analog_max is not None else None

    @digital_max.setter
    def digital_max(self, value):
        if value < 0 or value > 0xFFFF:
            raise ValueError("digital_max must be between 0 and 0xFFFF")
        self._digital_max = int(value)

    @property
    def shot_noise(self) -> bool:
        """when True, apply Poisson noise to the data instead of rounding"""
        return self._shot_noise

    @shot_noise.setter
    def shot_noise(self, value: bool):
        self._shot_noise = bool(value)

    @property
    def gaussian_noise_std(self) -> float:
        return self._gaussian_noise_std

    @gaussian_noise_std.setter
    def gaussian_noise_std(self, value: float):
        self._gaussian_noise_std = float(value)


class Camera(ADCProcessor):
    """Wraps any 2-D image source as a camera.

    The camera adds shot noise,

    To implement the camera interface, in addition to the Detector interface,
    we must implement left,right,top, and bottom.
    In addition, the data should be returned as uint16.
    Conversion to uint16 is implemented in the ADCProcessor base class.
    """

    def __init__(
        self,
        source: Detector,
        shape: Optional[Sequence[int]] = None,
        pos: Optional[Sequence[int]] = None,
        exposure: Quantity[u.ms] = 1 * u.ms,
        **kwargs,
    ):
        """
        Args:
            source (Detector): The source detector to be wrapped.
            shape (Optional[Sequence[int]]): The shape of the image data to be captured.
            pos (Optional[Sequence[int]]): The position on the source from where the image is captured.
            exposure (Quantity[u.ms]): The simulated exposure time. If the exposure time is changed later,
                the source data is scaled accordingly.
            **kwargs: Additional keyword arguments to be passed to the Detector base class.

            TODO: move left-right-top-bottom to CropProcessor.
            Expose the properties of CropProcessor as properties of MockCamera automatically by copying from __dict__?
        """
        self._crop = CropProcessor(source, shape=shape, pos=pos)
        super().__init__(source=self._crop, **kwargs)
        self._base_exposure = exposure
        self._exposure = exposure

    @property
    def left(self) -> int:
        """left (int): Horizontal start position of the ROI."""
        return self._crop.pos[1]

    @left.setter
    def left(self, value: int):
        self._crop.pos = (self._crop.pos[0], value)

    @property
    def right(self) -> int:
        """right (int): Horizontal end position of the ROI."""
        return self.left + self.width

    @property
    def top(self) -> int:
        """top (int): Vertical start position of the ROI."""
        return self._crop.pos[0]

    @top.setter
    def top(self, value: int):
        self._crop.pos = (value, self._crop.pos[1])

    @property
    def bottom(self) -> int:
        """bottom (int): Vertical end position of the ROI."""
        return self.top + self.height

    @property
    def height(self) -> int:
        """Height of the ROI."""
        return self._crop.data_shape[0]

    @height.setter
    def height(self, value: int):
        self._crop.data_shape = (value, self._crop.data_shape[1])

    @property
    def width(self) -> int:
        """Width of the ROI."""
        return self._crop.data_shape[1]

    @width.setter
    def width(self, value: int):
        self._crop.data_shape = (self._crop.data_shape[0], value)

    @property
    def data_shape(self):
        return self._crop.data_shape

    @data_shape.setter
    def data_shape(self, value):
        self._crop.data_shape = value

    @property
    def exposure(self) -> Quantity[u.ms]:
        return self._exposure

    @exposure.setter
    def exposure(self, value: Quantity[u.ms]):
        if value < 0.0:
            raise ValueError("Exposure time must be non-negative")
        self._exposure = value.to(u.ms)
        self._scale = self._exposure / self._base_exposure


class Stage(Actuator):
    """
    Mimics a single-axis stage actuator.
    """

    def __init__(self, axis: str, step_size: Quantity[u.um]):
        """
        Args:
            axis (str): The axis of the stage. Suggested usage is single characters (i.e. 'x', 'y', 'z').
            step_size (Quantity[u.um]): The step size of the stage along `axis` (in micrometers).
        """
        super().__init__(duration=0 * u.ms, latency=0 * u.ms)
        self._position = 0.0 * u.um
        self._axis = axis
        self._step_size = step_size.to(u.um)

    @property
    def axis(self) -> str:
        return self._axis

    @property
    def step_size(self) -> Quantity[u.um]:
        return self._step_size

    @property
    def position(self) -> Quantity[u.um]:
        return self._position

    @position.setter
    def position(self, value: Quantity[u.um]):
        self._position = self.step_size * np.round(value.to(u.um) / self.step_size)

    def home(self):
        self._position = 0.0 * u.um


class XYStage(Actuator):
    """
    Mimics an XY stage actuator
    """

    def __init__(self, step_size_x: Quantity[u.um], step_size_y: Quantity[u.um]):
        """

        Args:
            step_size_x (Quantity[u.um]): The step size in the x-direction.
            step_size_y (Quantity[u.um]): The step size in the y-direction.
        """
        super().__init__(duration=0 * u.ms, latency=0 * u.ms)
        self._step_size_x = step_size_x.to(u.um)
        self._step_size_y = step_size_y.to(u.um)
        self._y = 0.0 * u.um
        self._x = 0.0 * u.um

    @property
    def step_size_x(self) -> Quantity[u.um]:
        return self._step_size_x

    @property
    def step_size_y(self) -> Quantity[u.um]:
        return self._step_size_y

    @property
    def x(self) -> Quantity[u.um]:
        return self._x

    @x.setter
    def x(self, value: Quantity[u.um]):
        self._x = self.step_size_x * np.round(value.to(u.um) / self.step_size_x)

    @property
    def y(self) -> Quantity[u.um]:
        return self._y

    @y.setter
    def y(self, value: Quantity[u.um]):
        self._y = self.step_size_y * np.round(value.to(u.um) / self.step_size_y)

    def home(self):
        self._x = 0.0 * u.um
        self._y = 0.0 * u.um


class LinearStage(Actuator):
    """
    Mimics a linear translation stage
    """

    def __init__(self, step_size: Quantity[u.um]):
        """

        Args:
            step_size (Quantity[u.um]): The step size of the stage.
        """
        super().__init__(duration=0 * u.ms, latency=0 * u.ms)
        self._step_size = step_size.to(u.um)
        self._position = 0.0 * u.um

    @property
    def step_size(self) -> Quantity[u.um]:
        return self._step_size

    @property
    def position(self) -> Quantity[u.um]:
        return self._position

    @position.setter
    def position(self, value: Quantity[u.um]):
        self._position = self._step_size * np.round(unitless(value / self._step_size))

    def home(self):
        self._position = 0.0 * u.um


class Shutter(Processor):
    """
    A mock version of a shutter.
    When open, passes through the input field
    When closed, passes through the input field * 0.0
    """

    def __init__(self, source: Detector):
        super().__init__(source, multi_threaded=False)
        self._open = True

    @property
    def open(self) -> bool:
        return self._open

    @open.setter
    def open(self, value: bool):
        self._open = value

    def _fetch(self, source: np.ndarray) -> np.ndarray:  # noqa
        return source if self._open else 0.0 * source


class GaussianNoise(Processor):
    """Adds gaussian noise of a specified standard deviation to the signal
    Args:
        source (Detector): The source detector object to process the data from.
        std (float): The standard deviation of the gaussian noise.
        multi_threaded: Whether to perform processing in a worker thread.
    """

    def __init__(self, source: Detector, std: float, multi_threaded: bool = False):
        super().__init__(source, multi_threaded=multi_threaded)
        self._std = std

    @property
    def std(self) -> float:
        return self._std

    @std.setter
    def std(self, value: float):
        if value < 0.0:
            raise ValueError("Standard deviation must be non-negative")
        self._std = float(value)

    def _fetch(self, data: np.ndarray) -> np.ndarray:  # noqa
        """
        Args:
            data (ndarray): source data

        Returns: the out array containing the image with added noise.

        """
        return data + np.random.normal(0.0, self.std, data.shape)



import numpy as np

from .utilities import analyze_phase_stepping, WFSResult, DummyProgressBar
from ..core import Detector, PhaseSLM


class StepwiseSequential:
    """
    Class definition for stepwise sequential algorithm for wavefront shaping, as described by Vellekoop [1] [2].

    TODO: enable low-res pre-optimization (as in Vellekoop & Mosk 2007)
    TODO: modulate segments in pupil (circle) only

    [1]: Vellekoop, I. M., & Mosk, A. P. (2007). Focusing coherent light through opaque strongly scattering media.
    Optics Letters, 32(16), 2309-2311.
    [2]: Ivo M. Vellekoop, "Feedback-based wavefront shaping," Opt. Express 23, 12189-12206 (2015)
    """

    def __init__(
        self,
        feedback: Detector,
        slm: PhaseSLM,
        phase_steps: int = 4,
        n_x: int = 4,
        n_y: int = None,
    ):
        """
        This algorithm systematically modifies the phase pattern of each SLM element and measures the resulting
        feedback.

        Args:
            feedback (Detector): Source of feedback
            slm (PhaseSLM): The spatial light modulator
            phase_steps (int): The number of phase steps.
            n_x (int): Number of SLM elements in x direction
            n_y (int): Number of SLM elements in y direction
        """
        self._n_x = n_x
        self._n_y = n_x if n_y is None else n_y
        self._phase_steps = phase_steps
        self.slm = slm
        self.feedback = feedback

    def execute(self, progress_bar=DummyProgressBar()) -> WFSResult:
        """Executes the StepwiseSequential algorithm, computing the transmission matrix of the sample

        Returns:
            WFSResult: An object containing the computed transmission matrix and statistics.
        """
        phase_pattern = np.zeros((self.n_y, self.n_x), "float32")
        measurements = np.zeros((self.n_y, self.n_x, self.phase_steps, *self.feedback.data_shape))
        progress_bar.count = self.n_x * self.n_y

        for y in range(self.n_y):
            for x in range(self.n_x):
                for p in range(self.phase_steps):
                    phase_pattern[y, x] = p * 2 * np.pi / self.phase_steps
                    self.slm.set_phases(phase_pattern)
                    self.feedback.trigger(out=measurements[y, x, p, ...])
                phase_pattern[y, x] = 0
                progress_bar.update()

        self.feedback.wait()
        return analyze_phase_stepping(measurements, axis=2)

    @property
    def n_x(self) -> int:
        """The number of SLM elements in the x direction."""
        return self._n_x

    @n_x.setter
    def n_x(self, value):
        self._n_x = value

    @property
    def n_y(self) -> int:
        """The number of SLM elements in the y direction."""
        return self._n_y

    @n_y.setter
    def n_y(self, value):
        self._n_y = value

    @property
    def phase_steps(self) -> int:
        """The number of phase steps used for each segment."""
        return self._phase_steps

    @phase_steps.setter
    def phase_steps(self, value):
        self._phase_steps = value



import astropy.units as u
import numpy as np
from astropy.units import Quantity

from openwfs.processors import HDRCamera
from openwfs.simulation import StaticSource, SLM, Stage, Camera


def test_mock_slm():
    # TODO: at some point, merge this with the test_slm.py file
    # create a mock SLM object with a given shape
    slm = SLM(shape=(480, 640))

    # clear the slm
    slm.set_phases(0)

    # test if all elements of slm.phases are 0
    assert np.all(slm.phases.read() == 0)

    # set a pattern, don't update yet
    pattern = np.random.uniform(0.0, 1.9 * np.pi, size=(48, 64))
    slm.set_phases(pattern, update=False)

    # test if all elements of slm.phases are still 0
    assert np.all(slm.phases.read() == 0)

    # update the slm, and read back
    slm.update()
    scaled_pattern = np.repeat(np.repeat(pattern, 10, axis=0), 10, axis=1)
    assert np.allclose(slm.phases.read(), scaled_pattern, atol=1.0 * np.pi / 256)


def test_single_stage():
    stage = Stage("x", Quantity(10, u.um))
    assert stage.position == 0
    assert stage.axis == "x"

    stage.position = Quantity(5, u.um)
    assert stage.position == Quantity(0, u.um)

    stage.position = Quantity(15, u.um)
    assert stage.position == Quantity(20, u.um)

    stage.position = Quantity(-5, u.um)
    assert stage.position == Quantity(0, u.um)

    stage.position = Quantity(-15, u.um)
    assert stage.position == Quantity(-20, u.um)


def test_mock_camera():
    """Tests a simulated camera and HDR processing"""

    # create a mock camera object with a static image
    img = np.random.uniform(0, 2.0, size=(48, 64))
    img[0, 0] = 0.9813272622363753
    src = StaticSource(img, pixel_size=50 * u.nm)
    cam = Camera(
        src,
        digital_max=0xFF,
        analog_max=3.0,
        amplifier_bias=10,
        shot_noise=False,
        gaussian_noise_std=0.0,
        exposure=0.1 * u.ms,
    )

    # read the image from the camera, there should be an offset and small quantization error
    assert np.allclose(cam.read(), (255 / 3.0) * img + 10, atol=0.5)

    # change the exposure time and read again.
    # the image should be the same, but the exposure time should be different, and some pixels saturated
    cam.exposure = cam.exposure * 10
    assert np.allclose(cam.read(), np.minimum((255 / 3.0 * 10) * img + 10, 0xFF), atol=0.5)

    # perform HDR imaging with the camera, this should return the image without saturation
    # start with a trivial test, only one exposure, no background subtraction
    hdr = HDRCamera(cam, exposure_factors=(1.0,), background=0, saturation_threshold=255)
    assert np.allclose(hdr.read(), cam.read())

    hdr = HDRCamera(cam, exposure_factors=(1.0, 0.1, 0.01), background=10, saturation_threshold=250)
    frame = hdr.read()

    atol = (3 * 0.5) / 1.11 * 10  # worst case rounding error
    assert np.allclose(frame, (255 / 3.0 * 10) * img, atol=atol)



from . import algorithms
from . import core
from . import devices
from . import processors
from . import simulation
from . import utilities
from .core import Detector, Device, Actuator, Processor, PhaseSLM
from ._version import __version__, __git_revision__


def version():
    """Return the version string of the installed OpenWFS package.

    This includes __version__ and __git_revision__.
    __version__ is the version of the installed package, as reported by the package metadata.
    __git_revision__ is a string composed of the latest tag found in the code branch + the number of commits since that tag + the hash of the current commit.
    """
    return f"OpenWFS {__version__} ({__git_revision__})"



import importlib
import sys
from types import ModuleType


class _MockModule(ModuleType):
    """A mock object that replaces a missing module, or a function / class from that module.

    This class makes it possible to run openwfs without all optional dependencies installed.
    If `safe_import` tries to load a package that is not installed, it creates a `MockModule`
    as placeholder for that package. The openwfs subpackage that tried to load this package
    will then load normally (also so that readthedocs can parse the docstrings).
    Only when a function from the missing package is called, an error is raised (see `__call__`)."""

    def __init__(self, module_name: str, extra_name: str):
        super().__init__(module_name)
        self.module_name = module_name
        self.extra_name = extra_name
        self.message = f"""Module {self.module_name} is not installed.
            To install, use:
                pip install openwfs[{self.extra_name}]
            or
                pip install openwfs[all]
            """

    def __getattr__(self, name):
        """Mimicks accessing a submodule, function or class in the module.

        This just returns another MockModule object. Only when this object is called (e.g. `glfw.init()`),
        an error is raised.
        """
        if name.startswith("_"):
            return super().__getattr__(name)
        else:
            return _MockModule(self.module_name, self.extra_name)

    def __call__(self, *args, **kwargs):
        raise ModuleNotFoundError(self.message)


def is_loaded(module) -> bool:
    """Helper function to check if a module is a real loaded module or a mock module."""
    return not isinstance(module, _MockModule)


def safe_import(module_name: str, extra_name: str):
    """Tries to import a module.

    If the import files (usually because the module is not installed),
    this function registers a mock module so that importing can still continue without
    giving an error. This is essential for loading openwfs if optional dependencies are missing.
    This is also important for `readthedocs`, where not all dependenceis can be installed."""
    try:
        importlib.import_module(module_name)
    except (ModuleNotFoundError, AttributeError):
        sys.modules[module_name] = _MockModule(module_name, extra_name)


safe_import("harvesters", "genicam")
safe_import("harvesters.core", "genicam")
safe_import("nidaqmx", "nidaq")
safe_import("OpenGL", "opengl")
safe_import("glfw", "opengl")
safe_import("zaber_motion", "zaber")
safe_import("serial", "zaber")

from .camera import Camera
from .galvo_scanner import ScanningMicroscope, Axis
from .nidaq_gain import Gain
from . import slm
from .slm import SLM
from .zaber_stage import ZaberXYStage, ZaberLinearStage



from . import microscope
from . import mockdevices
from . import slm
from . import transmission

from .microscope import Microscope
from .mockdevices import (
    Stage,
    XYStage,
    LinearStage,
    StaticSource,
    Camera,
    ADCProcessor,
    Shutter,
    NoiseSource,
)
from .slm import SLM, PhaseToField
from .transmission import SimulatedWFS



from typing import Tuple, Union, Optional, Dict

import numpy as np
from astropy import units as u
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import hsv_to_rgb
from numpy import ndarray as nd

from .core import Detector
from .utilities import get_extent


# TODO: needs review and documentation. Remove single-use functions, simplify code.


def grab_and_show(cam: Detector, axis=None):
    return imshow(cam.read(), axis=axis)


def imshow(data, axis=None):
    extent = get_extent(data)
    e0 = scale_prefix(extent[0])
    e1 = scale_prefix(extent[1])
    if axis is None:
        plt.imshow(data, extent=(0.0, e1.value, 0.0, e0.value), cmap="gray")
        plt.colorbar()
        axis = plt.gca()
    else:
        axis.imshow(data, extent=(0.0, e1.value, 0.0, e0.value), cmap="gray")
    plt.ylabel(e0.unit.to_string())
    plt.xlabel(e1.unit.to_string())
    plt.show(block=False)
    plt.pause(0.1)
    return axis


def scale_prefix(value: u.Quantity) -> u.Quantity:
    """Scale a quantity to the most appropriate prefix unit."""
    if value.unit.physical_type == "length":
        if value < 100 * u.nm:
            return value.to(u.nm)
        if value < 100 * u.um:
            return value.to(u.um)
        if value < 100 * u.mm:
            return value.to(u.mm)
        else:
            return value.to(u.m)
    elif value.unit.physical_type == "time":
        if value < 100 * u.ns:
            return value.to(u.ns)
        if value < 100 * u.us:
            return value.to(u.us)
        if value < 100 * u.ms:
            return value.to(u.ms)
        else:
            return value.to(u.s)
    else:
        return value


def slope_step(a: nd, width: Union[nd, float]) -> nd:
    """
    A sloped step function from 0 to 1.

    Args:
        a: Input array
        width: width of the sloped step.

    Returns:
        An array the size of a, with the result of the sloped step function.
    """
    return (a >= width) + a / width * (0 < a) * (a < width)


def linear_blend(a: nd, b: nd, blend: Union[nd, float]) -> nd:
    """
    Return a linear, element-wise blend between two arrays a and b.

    Args:
        a: Input array a.
        b: Input array b.
        blend: Blend factor. Value of 1.0 -> return a. Value of 0.0 -> return b.

    Returns:
        A linear combination of a and b, corresponding to the blend factor. a*blend + b*(1-blend)
    """
    return a * blend + b * (1 - blend)


def complex_to_rgb(array: nd, scale: Optional[Union[float, nd]] = None, axis: int = 2) -> nd:
    """
    Generate RGB color values to represent values of a complex array.

    The complex values are mapped to HSV colorspace and then converted to RGB. Hue represents phase and Value represents
    amplitude. Saturation is set to 1.

    Args:
        array: Array to create RGB values for.
        scale: Scaling factor for the array values. When None, scale = 1/max(abs(array)) is used.
        axis: Array axis to use for the RGB dimension.

    Returns:
        An RGB array representing the complex input array.
    """
    if scale is None:
        scale = 1 / np.max(abs(array))
    h = np.expand_dims(np.angle(array) / (2 * np.pi) + 0.5, axis=axis)
    s = np.ones_like(h)
    v = np.expand_dims(np.abs(array) * scale, axis=axis).clip(min=0, max=1)
    hsv = np.concatenate((h, s, v), axis=axis)
    rgb = hsv_to_rgb(hsv)
    return rgb


def plot_field(array, scale: Optional[Union[float, nd]] = None, imshow_kwargs: Optional[Dict] = None):
    """
    Plot a complex array as an RGB image.

    The phase is represented by the hue, and the magnitude by the value, i.e. black = zero, brightness shows amplitude,
    and the colors represent the phase.

    Args:
        array(ndarray): complex array to be plotted.
        scale(float): scaling factor for the magnitude. The final value is clipped to the range [0, 1].
        imshow_kwargs: Keyword arguments for matplotlib's imshow.
    """
    if imshow_kwargs is None:
        imshow_kwargs = {}
    rgb = complex_to_rgb(array, scale)
    plt.imshow(rgb, **imshow_kwargs)


def plot_scatter_field(x, y, array, scale, scatter_kwargs=None):
    """
    Plot complex scattered data as RGB values.
    """
    if scatter_kwargs is None:
        scatter_kwargs = {"s": 80}
    rgb = complex_to_rgb(array, scale, axis=1)
    plt.scatter(x, y, c=rgb, **scatter_kwargs)


def complex_colorbar(scale, width_inverse: int = 15):
    """
    Create an rgb colorbar for complex numbers and return its Axes handle.
    """
    amp = np.linspace(0, 1.01, 10).reshape((1, -1))
    phase = np.linspace(0, 249 / 250 * 2 * np.pi, 250).reshape(-1, 1) - np.pi
    z = amp * np.exp(1j * phase)
    rgb = complex_to_rgb(z, 1)
    ax = plt.subplot(1, width_inverse, width_inverse)
    plt.imshow(rgb, aspect="auto", extent=(0, scale, -np.pi, np.pi))

    # Ticks and labels
    ax.set_yticks((-np.pi, -np.pi / 2, 0, np.pi / 2, np.pi), ("$-\\pi$", "$-\\pi/2$", "0", "$\\pi/2$", "$\\pi$"))
    ax.set_xlabel("amp.")
    ax.set_ylabel("phase (rad)")
    ax.yaxis.tick_right()
    ax.yaxis.set_label_position("right")
    return ax


def complex_colorwheel(
    ax: Axes = None,
    shape: Tuple[int, int] = (100, 100),
    imshow_kwargs: dict = {},
    arrow_props: dict = {},
    text_kwargs: dict = {},
    amplitude_str: str = "A",
    phase_str: str = "$\\phi$",
):
    """
    Create an rgb image for a colorwheel representing the complex unit circle.
    TODO: needs review

    Args:
        ax: Matplotlib Axes.
        shape: Number of pixels in each dimension.
        imshow_kwargs: Keyword arguments for matplotlib's imshow.
        arrow_props: Keyword arguments for the arrows.
        text_kwargs: Keyword arguments for the text labels.
        amplitude_str: Text label for the amplitude arrow.
        phase_str: Text label for the phase arrow.

    Returns:
        rgb_wheel: rgb image of the colorwheel.
    """
    if ax is None:
        ax = plt.gca()

    x = np.linspace(-1, 1, shape[1]).reshape(1, -1)
    y = np.linspace(-1, 1, shape[0]).reshape(-1, 1)
    z = x + 1j * y
    rgb = complex_to_rgb(z, scale=1)
    step_width = 1.5 / shape[1]
    blend = np.expand_dims(slope_step(1 - np.abs(z) - step_width, width=step_width), axis=2)
    rgba_wheel = np.concatenate((rgb, blend), axis=2)
    ax.imshow(rgba_wheel, extent=(-1, 1, -1, 1), **imshow_kwargs)

    # Add arrows with annotations
    ax.annotate(
        "",
        xy=(-0.98 / np.sqrt(2),) * 2,
        xytext=(0, 0),
        arrowprops={"color": "white", "width": 1.8, "headwidth": 5.0, "headlength": 6.0, **arrow_props},
    )
    ax.text(**{"x": -0.4, "y": -0.8, "s": amplitude_str, "color": "white", "fontsize": 15, **text_kwargs})
    ax.annotate(
        "",
        xy=(0, 0.9),
        xytext=(0.9, 0),
        arrowprops={
            "connectionstyle": "arc3,rad=0.4",
            "color": "white",
            "width": 1.8,
            "headwidth": 5.0,
            "headlength": 6.0,
            **arrow_props,
        },
    )
    ax.text(**{"x": 0.1, "y": 0.5, "s": phase_str, "color": "white", "fontsize": 15, **text_kwargs})

    # Hide axes spines and ticks
    ax.set_xticks([])
    ax.set_yticks([])
    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)



from typing import Optional, List

import numpy as np
from numpy import ndarray as nd

from .utilities import analyze_phase_stepping, WFSResult, DummyProgressBar
from ..core import Detector, PhaseSLM


class DualReference:
    """A generic iterative dual reference WFS algorithm, which can use a custom set of basis functions.

    This algorithm is adapted from [Tao2017]_, with the addition of the ability to use custom basis functions and
    specify the number of iterations.

    In this algorithm, the SLM pixels are divided into two groups:
    A and B, as indicated by the boolean group_mask argument.
    The algorithm first keeps the pixels in group B fixed, and displays a sequence on patterns on the pixels of group A.
    It uses these measurements to construct an optimized wavefront that is displayed on the pixels of group A.
    This process is then repeated for the pixels of group B, now using the *optimized* wavefront on group A as
    reference. Optionally, the process can be repeated for a number of iterations, which each iteration using the current correction
    pattern as a reference. This makes this algorithm suitable for non-linear feedback, such as multi-photon
    excitation fluorescence [Osnabrugge2019]_.

    This algorithm assumes a phase-only SLM. Hence, the input modes are defined by passing the corresponding phase
    patterns (in radians) as input argument.

    References
    ----------
    .. [Tao2017] X. Tao, T. Lam, B. Zhu, et al., “Three-dimensional focusing through scattering media using conjugate
            adaptive optics with remote focusing (CAORF),” Opt. Express 25, 10368–10383 (2017).

    .. [Osnabrugge2019] Gerwin Osnabrugge, Lyubov V. Amitonova, and Ivo M. Vellekoop. "Blind focusing through strongly
            scattering media using wavefront shaping with nonlinear feedback", Optics Express, 27(8):11673–11688, 2019.
            https://opg.optica.org/oe/ abstract.cfm?uri=oe-27-8-1167

    """

    def __init__(
        self,
        *,
        feedback: Detector,
        slm: PhaseSLM,
        phase_patterns: Optional[tuple[nd, nd]],
        group_mask: nd,
        amplitude: nd = 1.0,
        phase_steps: int = 4,
        iterations: int = 2,
        optimized_reference: Optional[bool] = None,
    ):
        """
        Args:
            feedback: The feedback source, usually a detector that provides measurement data.
            slm: Spatial light modulator object.
            phase_patterns:
                A tuple of two 3D arrays, containing the phase patterns for group A and group B, respectively.
                The 3D arrays have shape ``(pattern_count, height, width)``.
                The number of phase patterns in A and B may be different.
            amplitude:
                2D amplitude distribution on the SLM, should have shape ``(height, width)``.
            group_mask: A 2D bool array of that defines the pixels used by group A with False and elements used by
                group B with True, should have shape ``(height, width)``.
            phase_steps: The number of phase steps for each mode (default is 4). Depending on the type of
                non-linear feedback and the SNR, more might be required.
            iterations: Number of times to optimize a mode set, e.g. when iterations = 5, the measurements are
                A, B, A, B, A.
            optimized_reference:
                When `True`, during each iteration the other half of the SLM displays the optimized pattern so far (as in [1]).
                When `False`, the algorithm optimizes A with a flat wavefront on B,
                and then optimizes B with a flat wavefront on A.
                This mode also allows for multi-target optimization,
                where the algorithm optimizes multiple targets in parallel.
                The two halves are then combined (stitched) to form the full transmission matrix.
                In this mode, it is essential that both A and B include a flat wavefront as mode 0.
                The measurement for mode A0 and for B0 both give contain relative phase between group A and B,
                so there is a slight redundancy.
                These two measurements are combined to find the final phase for stitching.
                When set to `None` (default), the algorithm uses True if there is a single target,
                and False if there are multiple targets.

        [1]: X. Tao, T. Lam, B. Zhu, et al., “Three-dimensional focusing through scattering media using conjugate adaptive
        optics with remote focusing (CAORF),” Opt. Express 25, 10368–10383 (2017).
        """
        if optimized_reference is None:  # 'auto' mode
            optimized_reference = np.prod(feedback.data_shape) == 1
        elif optimized_reference and np.prod(feedback.data_shape) != 1:
            raise ValueError("In optimized_reference mode, only scalar (single target) feedback signals can be used.")
        if iterations < 2:
            raise ValueError("The number of iterations must be at least 2.")
        if not optimized_reference and iterations != 2:
            raise ValueError("When not using an optimized reference, the number of iterations must be 2.")

        self.slm = slm
        self.feedback = feedback
        self.phase_steps = phase_steps
        self.optimized_reference = optimized_reference
        self.iterations = iterations
        self._phase_patterns = None
        self._gram = None
        self._shape = group_mask.shape
        mask = group_mask.astype(bool)
        self.masks = (
            ~mask,
            mask,
        )  # self.masks[0] is True for group A, self.masks[1] is True for group B
        self._amplitude = 1.0
        self.amplitude = amplitude
        self.phase_patterns = phase_patterns

    @property
    def amplitude(self) -> Optional[nd]:
        return self._amplitude

    @amplitude.setter
    def amplitude(self, value: np.ndarray):
        if not np.isscalar(value) and value.shape != self._shape:
            raise ValueError("The amplitude and group mask must all have the same shape.")

        self._amplitude = value

    @property
    def phase_patterns(self) -> tuple[nd, nd]:
        return self._phase_patterns

    @phase_patterns.setter
    def phase_patterns(self, value: np.ndarray):
        """Sets the phase patterns for group A and group B. This also updates the conjugate modes."""
        self._zero_indices = [0, 0]
        self._phase_patterns = [None, None]
        self._cobasis = [None, None]
        self._gram = [None, None]

        for side in range(2):
            patterns = value[side]
            mask = self.masks[side]
            if patterns[0].shape != self._shape:
                raise ValueError("The phase patterns and group mask must all have the same shape.")
            if not self.optimized_reference:
                # find the modes in A and B that correspond to flat wavefronts with phase 0
                try:
                    self._zero_indices[side] = next(i for i, p in enumerate(patterns) if np.allclose(p, 0))
                except StopIteration:
                    raise "For multi-target optimization, the both sets must contain a flat wavefront with phase 0."

            self._phase_patterns[side] = patterns.astype(np.float32)

            # Computes the cobasis
            # As a basis matrix B is full rank, the cobasis is equivalent to the Moore-Penrose pseudo-inverse B⁺.
            # B⁺ = (B^* B)⁻¹ B^*
            # Where B is the basis matrix, ^* denotes the conjugate transpose, and ^(-1)
            # denotes the matrix inverse. Note: we store the cobasis in transposed form,
            # with shape (mode, y, x) to allow for easy
            # multiplication with the transmission matrix t_ba.
            basis = self.amplitude * mask * np.exp(1j * patterns)
            basis /= np.linalg.norm(basis, axis=(1, 2), keepdims=True)
            gram = np.tensordot(basis, basis.conj(), axes=((1, 2), (1, 2)))  # inner product (contracts x and y axis)
            self._cobasis[side] = np.tensordot(np.linalg.inv(gram), basis.conj(), 1)
            self._gram[side] = gram

    @property
    def cobasis(self) -> tuple[nd, nd]:
        """
        The cobasis corresponding to the given basis.

        Note: The cobasis is stored in transposed form, with shape = (mode_count, height, width)
        """
        return tuple(self._cobasis)

    @property
    def gram(self) -> tuple[nd, nd]:
        """
        The Gram matrix corresponding to the given basis (i.e. phase pattern and amplitude profile).
        """
        return tuple(self._gram)

    def execute(self, *, capture_intermediate_results: bool = False, progress_bar=DummyProgressBar()) -> WFSResult:
        """
        Executes the blind focusing dual reference algorithm and compute the SLM transmission matrix.
            capture_intermediate_results: When True, measures the feedback from the optimized wavefront after each iteration.
                This can be useful to determine how many iterations are needed to converge to an optimal pattern.
                This data is stored as the 'intermediate_results' field in the results
            progress_bar: Optional progress bar object. Following the convention for tqdm progress bars,
                this object should have a `total` attribute and an `update()` function.

        Returns:
            WFSResult: An object containing the computed SLM transmission matrix and related data. The amplitude profile
                of each mode is assumed to be 1. If a different amplitude profile is desired, this can be obtained by
                multiplying that amplitude profile with this transmission matrix.
        """

        # Current estimate of the transmission matrix (start with all 0)
        ref_phases = np.zeros(self._shape)

        # Initialize storage lists
        results_all: List[Optional[WFSResult]] = [None] * self.iterations  # List to store all results
        intermediate_results = np.zeros(self.iterations)  # List to store feedback from full patterns

        # Prepare progress bar
        progress_bar.total = (
            np.ceil(self.iterations / 2) * self.phase_patterns[0].shape[2]
            + np.floor(self.iterations / 2) * self.phase_patterns[1].shape[2]
        )

        # Switch the phase sets back and forth multiple times
        for it in range(self.iterations):
            side = it % 2  # pick set A or B for phase stepping
            side_mask = self.masks[side]
            # Perform WFS experiment on one side, keeping the other side sized at the ref_phases
            results_all[it] = self._single_side_experiment(
                mod_phases=self.phase_patterns[side],
                ref_phases=ref_phases,
                mod_mask=side_mask,
                progress_bar=progress_bar,
            )

            # Compute transmission matrix for the current side and update
            # estimated transmission matrix

            if self.optimized_reference:
                # use the best estimate so far to construct an optimized reference
                # TODO: see if the squeeze can be removed
                t_this_side = self._compute_t_set(results_all[it].t, side).squeeze()
                ref_phases[self.masks[side]] = -np.angle(t_this_side[self.masks[side]])

            # Try full pattern
            if capture_intermediate_results:
                self.slm.set_phases(ref_phases)
                intermediate_results[it] = self.feedback.read()

        if self.iterations % 2 == 0:
            t_side_0 = results_all[-2].t
            t_side_1 = results_all[-1].t
        else:
            t_side_0 = results_all[-1].t
            t_side_1 = results_all[-2].t

        if self.optimized_reference:
            factor = 1.0
        else:
            # when not using optimized reference, we need to stitch the
            # two halves of the wavefront together. For that, we need the
            # relative phase between the two sides, which we extract from
            # the measurements of the flat wavefronts.
            relative = t_side_0[..., self._zero_indices[0]] + np.conjugate(t_side_1[..., self._zero_indices[1]])
            factor = np.expand_dims(relative / np.abs(relative), -1)

        t_full = self._compute_t_set(t_side_0, 0) + self._compute_t_set(factor * t_side_1, 1)

        # Compute average fidelity factors
        # subtract 1 from n, because both sets (usually) contain a flat wavefront,
        # so there is one redundant measurement
        result = WFSResult.combine(results_all[-2:])
        result.n = result.n - 1
        result.t = t_full

        # TODO: document the results_all attribute
        result.results_all = results_all
        result.intermediate_results = intermediate_results
        return result

    def _single_side_experiment(self, mod_phases: nd, ref_phases: nd, mod_mask: nd, progress_bar) -> WFSResult:
        """
        Conducts experiments on one part of the SLM.

        Args:
            mod_phases: 3D array containing the phase patterns of each mode.
                ``shape = mode_count × height × width``
            ref_phases: 2D array containing the reference phase pattern.
            mod_mask: 2D array containing a boolean mask, where True indicates the modulated part of the SLM.
            progress_bar: Progress bar object. Following the convention for tqdm progress bars,
                this object should have a `total` attribute and an `update()` function.

        Returns:
            WFSResult: An object containing the computed SLM transmission matrix and related data.
        """
        measurements = np.zeros((len(mod_phases), self.phase_steps, *self.feedback.data_shape))

        for m, modulated in enumerate(mod_phases):
            phases = ref_phases.copy()
            for p in range(self.phase_steps):
                phi = p * 2 * np.pi / self.phase_steps
                # set the modulated pixel values to the values corresponding to mode m and phase offset phi
                phases[mod_mask] = modulated[mod_mask] + phi
                self.slm.set_phases(phases)
                self.feedback.trigger(out=measurements[m, p, ...])
                progress_bar.update()

        self.feedback.wait()
        return analyze_phase_stepping(measurements, axis=1)

    def _compute_t_set(self, t, side) -> nd:
        """
        Compute the transmission matrix in SLM space from transmission matrix in input mode space.

        Equivalent to ``np.tensordot(t, cobasis[side], 1)``

        Args:
            t: transmission matrix in mode-index space. The first axis corresponds to the input modes.
        Returns:
            nd: The transmission matrix in SLM space. The last two axes correspond to SLM coordinates
        """
        return np.tensordot(t, self.cobasis[side], 1)



from . import patterns
from . import utilities
from .patterns import coordinate_range, disk, gaussian, tilt
from .utilities import (
    ExtentType,
    CoordinateType,
    unitless,
    get_pixel_size,
    set_pixel_size,
    Transform,
    project,
    place,
    set_extent,
    get_extent,
)



import astropy.units as u
import numpy as np
import pytest
import skimage as sk

from openwfs.processors import SingleRoi, select_roi, Roi, MultipleRoi
from openwfs.simulation.mockdevices import StaticSource


@pytest.mark.skip(
    reason="This is an interactive test: skip by default. TODO: actually test if the roi was " "selected correctly."
)
def test_croppers():
    img = sk.data.camera()
    src = StaticSource(img, pixel_size=50 * u.nm)
    roi = select_roi(src, "disk")
    assert roi.mask_type == "disk"


def test_single_roi_simple_case():
    data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    pixel_size = 1 * np.ones(2)
    mock_source = StaticSource(data, pixel_size=pixel_size)
    roi_processor = SingleRoi(mock_source, radius=np.sqrt(2))
    roi_processor.trigger()
    result = roi_processor.read()

    # Debugging: Print the mask and the selected region
    print("Mask:", roi_processor._rois[()]._mask)

    expected_value = np.mean(data[0:3, 0:3])  # Assuming this is how the ROI is defined
    assert np.isclose(
        result, expected_value
    ), f"ROI average value is incorrect. Expected: {expected_value}, Got: {result}"


def create_mock_source_with_data():
    data = np.arange(25).reshape(5, 5)
    return StaticSource(data, pixel_size=1 * u.um)


@pytest.mark.parametrize(
    "x, y, radius, expected_avg",
    [
        (2, 2, 1, 12),  # Center ROI in 5x5 matrix
        (0, 0, 0, 0),  # Top-left corner ROI in 5x5 matrix
    ],
)
def test_single_roi(x, y, radius, expected_avg):
    mock_source = create_mock_source_with_data()
    roi_processor = SingleRoi(mock_source, (y, x), radius)
    roi_processor.trigger()
    result = roi_processor.read()

    assert np.isclose(result, expected_avg), f"ROI average value is incorrect. Expected: {expected_avg}, Got: {result}"


def test_multiple_roi_simple_case():
    data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    pixel_size = 1 * np.ones(2)
    mock_source = StaticSource(data, pixel_size=pixel_size)

    rois = [
        Roi((1, 1), radius=0),
        Roi((2, 2), radius=0),
        Roi((1, 1), radius=1),
        Roi((0, 1), radius=0),
    ]

    roi_processor = MultipleRoi(mock_source, rois=rois)
    roi_processor.trigger()
    result = roi_processor.read()

    expected_values = [5, 9, 5, 2]
    assert all(
        np.isclose(r, e) for r, e in zip(result, expected_values)
    ), f"ROI average values are incorrect. Expected: {expected_values}, Got: {result}"
